import abc
import dataclasses
import logging
import sys
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import numpy as np
import torch.utils.data
import torch.distributed
from torch import Tensor

import ase.io
import ase.data
from franken.backbones import REGISTRY
from franken.backbones.utils import make_summary
from franken.data.distributed_sampler import SimpleUnevenDistributedSampler
from franken.rf.atomic_energies import AtomicEnergiesShift


logger = logging.getLogger("franken")


@torch.jit.script
class Configuration:
    """Container for a single configuration (molecule or crystal).

    The set of attributes which are non empty depends on the GNN backbone which the
    :class:`Configuration` object will be passed to.
    """

    def __init__(
        self,
        atom_pos: Tensor,
        atomic_numbers: Tensor,
        natoms: Tensor,
        node_attrs: Optional[Tensor] = None,
        edge_index: Optional[Tensor] = None,
        shifts: Optional[Tensor] = None,
        unit_shifts: Optional[Tensor] = None,
        cell: Optional[Tensor] = None,
        batch_ids: Optional[Tensor] = None,
        pbc: Optional[Tensor] = None,
    ):
        self.atom_pos = atom_pos
        self.atomic_numbers = atomic_numbers
        self.natoms = natoms
        self.node_attrs = node_attrs
        self.edge_index = edge_index
        self.shifts = shifts
        self.unit_shifts = unit_shifts
        self.cell = cell
        self.batch_ids = batch_ids
        self.pbc = pbc

    def to(
        self, device: torch.device | None = None, dtype: torch.dtype | None = None
    ) -> "Configuration":
        # optional-type refinement must be on local variables (torch.jit.script)
        node_attrs = self.node_attrs
        if node_attrs is not None:
            node_attrs = node_attrs.to(device=device, dtype=dtype)
        edge_index = self.edge_index
        if edge_index is not None:
            edge_index = edge_index.to(device=device, dtype=dtype)
        shifts = self.shifts
        if shifts is not None:
            shifts = shifts.to(device=device, dtype=dtype)
        unit_shifts = self.unit_shifts
        if unit_shifts is not None:
            unit_shifts = unit_shifts.to(device=device, dtype=dtype)
        cell = self.cell
        if cell is not None:
            cell = cell.to(device=device, dtype=dtype)
        batch_ids = self.batch_ids
        if batch_ids is not None:
            batch_ids = batch_ids.to(device=device, dtype=dtype)
        pbc = self.pbc
        if pbc is not None:
            pbc = pbc.to(device=device, dtype=dtype)
        return Configuration(
            atom_pos=self.atom_pos.to(device=device, dtype=dtype),
            atomic_numbers=self.atomic_numbers.to(device=device, dtype=dtype),
            natoms=self.natoms.to(device=device, dtype=dtype),
            node_attrs=node_attrs,
            edge_index=edge_index,
            shifts=shifts,
            unit_shifts=unit_shifts,
            cell=cell,
            batch_ids=batch_ids,
            pbc=pbc,
        )


@dataclasses.dataclass
class Target:
    """Container class for the target variables of a single configuration."""

    energy: Tensor
    forces: Optional[Tensor]

    def to(self, device=None, dtype=None) -> "Target":
        return Target(
            energy=self.energy.to(device=device, dtype=dtype),
            forces=(
                self.forces.to(device=device, dtype=dtype)
                if self.forces is not None
                else None
            ),
        )


class BaseAtomsDataset(torch.utils.data.Dataset, abc.ABC):
    """Base class for atom datasets.

    The data is loaded entirely into memory from the provided :attr:`dataset_dir` and the
    file-path specified in the :attr:`DATASET_REGISTRY`. Each dataset must be stored in a
    single `.extxyz` file which is read using `ase <https://wiki.fysik.dtu.dk/ase/>`_.

    Args:
        data_path (str or None): path to the '.extxyz' file to be loaded with `ase`.
            This can be None in which case no data will be loaded (useful for running MD).
        split (str): the split ('train', 'test', 'val', 'md') which will be used.  The 'md'
            split can be used to initialize an empty dataset to help running MD simulations.
        dataset_dir (str or Path): points to the base folder containing datasets,
            if set to `None` the root path of the repository is used.
        num_configurations: (optional int): maximum number of configurations to
            load. The default is to load all configurations in the file.
    """

    def __init__(
        self,
        data_path: str | Path | None,
        split: str,
        num_random_subsamples: int | None = None,
        subsample_rng: int | None = None,
    ):
        self.split = split
        self.data_path = data_path

        ase_atoms: list[ase.Atoms] = []
        if self.data_path is not None:
            read_ase_atoms = ase.io.read(self.data_path, index=":")
            if isinstance(read_ase_atoms, ase.Atoms):
                # workaround edge case of a single configuration
                ase_atoms = [read_ase_atoms]
            else:
                ase_atoms = read_ase_atoms

        if subsample_rng is not None:
            rng = np.random.default_rng(subsample_rng)
            rng.shuffle(ase_atoms)  # type: ignore

        if (num_random_subsamples is not None) and (
            1 <= num_random_subsamples <= len(ase_atoms)
        ):
            ase_atoms = ase_atoms[:num_random_subsamples]

        self.ase_atoms = ase_atoms
        self.atomic_energies_ = None
        self.energy_shifts_ = None
        self.mean_absolute_deviations_ = None

    def add_configuration(self, atoms: ase.Atoms) -> int:
        self.ase_atoms.append(atoms)
        return len(self.ase_atoms) - 1

    def __len__(self):
        return len(self.ase_atoms)

    def compute_average_atomic_energies(self) -> Dict[int, torch.Tensor]:
        """
        Function to compute the average interaction energy of each chemical element
        returns dictionary of E0s
        """
        len_train = len(self)
        zs = self.species
        len_zs = len(zs)
        A = torch.zeros((len_train, len_zs))
        B = torch.zeros(len_train)
        for i in range(len_train):
            atoms = self.ase_atoms[i]
            B[i] = atoms.get_potential_energy(apply_constraint=False)
            for j, z in enumerate(zs):
                A[i, j] = torch.count_nonzero(
                    torch.tensor(atoms.get_atomic_numbers() == z)
                )
        try:
            E0s = torch.linalg.lstsq(A, B, rcond=None)[0]
            atomic_energies_dict = {}
            for i, z in enumerate(zs):
                atomic_energies_dict[z] = E0s[i]
        except torch.linalg.LinAlgError:
            logging.warning(
                "Failed to compute atomic energies using least squares regression, using the same for all atoms"
            )
            atomic_energies_dict = {}
            for i, z in enumerate(zs):
                atomic_energies_dict[z] = B.sum() / A.sum()

        return atomic_energies_dict

    @property
    def atomic_energies(self):
        # average atomic energies
        if self.atomic_energies_ is None:
            if self.split != "train":
                raise RuntimeError(
                    f"Atomic energies should only be computed on the training set. Found: {self.split}."
                )
            self.atomic_energies_ = self.compute_average_atomic_energies()
        return self.atomic_energies_

    @property
    def energy_shifts(self):
        # precompute energy shifts based on atomic energies
        if self.energy_shifts_ is None:
            if self.split != "train":
                raise RuntimeError(
                    f"Energy shifts should only be computed for the training set. Found: {self.split}."
                )
            shifter = AtomicEnergiesShift(
                num_species=self.num_species, atomic_energies=self.atomic_energies
            )
            self.energy_shifts_ = [
                shifter(atoms.get_atomic_numbers()) for atoms in self.ase_atoms
            ]

        return self.energy_shifts_

    @property
    def mean_absolute_deviations(self):
        """Returns the mean absolute deviations of the potential energy and forces in the dataset."""
        if self.mean_absolute_deviations_ is None:
            energies = np.array(
                [a.get_potential_energy(apply_constraint=False) for a in self.ase_atoms]
            )
            forces = np.concatenate([a.get_forces() for a in self.ase_atoms], axis=0)
            mad_energy = np.mean(np.absolute(energies - np.mean(energies)))
            mad_forces = np.mean(
                np.absolute(forces - np.mean(forces, axis=0, keepdims=True))
            )
            self.mean_absolute_deviations_ = {
                "energy": float(mad_energy),
                "forces": float(mad_forces),
            }
        return self.mean_absolute_deviations_

    @property
    def species(self):
        _species = set()
        for a in self.ase_atoms:
            _species.update(a.get_atomic_numbers().tolist())
        return sorted(list(_species))

    @property
    def num_species(self):
        return len(self.species)

    @abc.abstractmethod
    def __getitem__(
        self, idx, no_targets: bool = False
    ) -> Union[Configuration, Tuple[Configuration, Target]]:
        pass

    @staticmethod
    def from_path(
        data_path: str | Path | None,
        split: str,
        gnn_backbone_id: str | None,
        num_random_subsamples: int | None = None,
        subsample_rng: int | None = None,
    ) -> "BaseAtomsDataset":
        """Factory method to initialize an atoms dataset with a specific GNN type.

        Datasets are tightly coupled to the GNN they will be used with: different
        kinds of GNNs require slightly different variations on the dataset class,
        for example see :class:`franken.datasets.atoms_dataset.MACEAtomsDataset` and
        :class:`franken.datasets.atoms_dataset.FairchemAtomsDataset`.
        This method will take care of instantiating the correct dataset class for
        the given GNN.

        Args:
            data_path (str or None): path to the '.extxyz' file to be loaded with `ase`.
                This can be None in which case no data will be loaded (useful for running MD).
            split (str): the split ('train', 'test', 'val', 'md') which will be used.  The 'md'
                split can be used to initialize an empty dataset to help running MD simulations.
            gnn_backbone_id (str): name of the GNN to be used with the dataset. If `None` is passed,
                we will initialize a simple dataset without any inter-atomic graph information.
            num_random_subsamples: (optional int): maximum number of configurations to
                load. The default is to load all configurations in the file.
            subsample_rng: (optional int): random number generator seed used while subsampling data.

        """
        if gnn_backbone_id is not None:
            if gnn_backbone_id not in REGISTRY.keys():
                raise NameError(
                    f"Unknown {gnn_backbone_id} GNN backbone, the current available backbones are\n{make_summary()}"
                )
            backbone_family = REGISTRY[gnn_backbone_id]["kind"].lower()
        else:
            backbone_family = None
        if backbone_family == "mace":
            try:
                from franken.data.mace import MACEAtomsDataset

                return MACEAtomsDataset(
                    data_path,
                    split,
                    num_random_subsamples,
                    subsample_rng,
                    gnn_backbone_id,
                    precompute=True,
                )
            except ImportError as import_err:
                print(
                    f"franken wasn't able to load {gnn_backbone_id}. Is {backbone_family} installed?",
                    file=sys.stderr,
                )
                raise import_err
        elif backbone_family == "fairchem":
            try:
                from franken.data.fairchem import FairchemAtomsDataset

                return FairchemAtomsDataset(
                    data_path,
                    split,
                    num_random_subsamples,
                    subsample_rng,
                    gnn_backbone_id,
                    precompute=True,
                )
            except ImportError as import_err:
                print(
                    f"franken wasn't able to load {gnn_backbone_id}. Is {backbone_family} installed?",
                    file=sys.stderr,
                )
                raise import_err
        elif backbone_family == "sevenn":
            try:
                from franken.data.sevenn import SevennAtomsDataset
            except ImportError as e:
                logger.error(
                    f"franken wasn't able to load {gnn_backbone_id}. Is {backbone_family} installed?",
                    exc_info=e,
                )
                raise
            return SevennAtomsDataset(
                data_path,
                split,
                num_random_subsamples,
                subsample_rng,
                gnn_backbone_id,
                precompute=True,
            )
        elif backbone_family is None:
            return SimpleAtomsDataset(
                data_path, split, num_random_subsamples, subsample_rng
            )
        else:
            raise ValueError(backbone_family)

    def get_dataloader(self, distributed: bool) -> torch.utils.data.DataLoader:
        """Get a dataloader corresponding to a :class:`~franken.datasets.atoms_dataset.BaseAtomsDataset`.

        The dataloader creation is specific to the problem: batch size is fixed to 1 and no extra workers are
        used since fetching the data is fast.
        If :attr:`distributed` is True the dataloader will use a :class:`torch.utils.data.DistributedSampler`
        to distribute the dataset's samples among available processes.

        Args:
            dset (BaseAtomsDataset): the dataset used by the dataloader
            split (str): the data-split ('train', 'test', 'val')
            distributed (bool): whether the dataloader should be distributed among available processes
        """

        def empty_collate_fn(batch):
            assert len(batch) == 1
            return batch[0]

        sampler = None
        if distributed and torch.distributed.is_initialized():
            sampler = SimpleUnevenDistributedSampler(self)
        elif distributed and not torch.distributed.is_initialized():
            logger.warning(
                "The distributed flag was set to True, but torch.distributed is not initialized. "
            )
        return torch.utils.data.DataLoader(
            dataset=self,
            batch_size=1,
            shuffle=False,
            sampler=sampler,
            drop_last=False,
            num_workers=0,
            collate_fn=empty_collate_fn,
        )


class SimpleAtomsDataset(BaseAtomsDataset):
    def __init__(
        self,
        data_path: str | Path | None,
        split: str,
        num_random_subsamples: int | None = None,
        subsample_rng: int | None = None,
    ):
        super().__init__(data_path, split, num_random_subsamples, subsample_rng)

    def __getitem__(self, idx, no_targets: bool = False):
        """Returns an array of (inputs, outputs) with inputs being a configuration
        and outputs being the target (energy and forces).
        Note: ONLY for the 'train' split, the energy_shift is removed from the target.
        """
        atoms = self.ase_atoms[idx]
        positions = atoms.get_positions()
        atomic_numbers = np.array(
            [ase.data.atomic_numbers[symbol] for symbol in atoms.symbols]
        )
        cell = np.array(atoms.get_cell())

        data = Configuration(
            atom_pos=torch.from_numpy(positions).to(dtype=torch.float32),
            atomic_numbers=torch.from_numpy(atomic_numbers).to(dtype=torch.int64),
            natoms=torch.tensor(len(atomic_numbers)).view(1),
            cell=torch.from_numpy(cell).to(dtype=torch.float32),
        )
        if no_targets:
            return data
        energy = torch.tensor(
            self.ase_atoms[idx].get_potential_energy(apply_constraint=False)
        )
        if self.split == "train":
            energy = energy - self.energy_shifts[idx]
        target = Target(
            energy=energy,
            forces=torch.Tensor(self.ase_atoms[idx].get_forces(apply_constraint=False)),
        )
        return data, target



from ase.io import read, write

if __name__ == "__main__":
    # Loading the dataset
    dataset = read("ML_AB_dataset_2.xyz", index=":", format="extxyz")
    # Saving new dataset
    write("ML_AB_dataset_2-val.xyz", dataset[473:])



import ase.io
from ase import Atoms
import numpy as np
import argparse

def create_periodic_cell_for_trajectory(trajectory_file, output_file, padding=100.0):
    # Load all frames from the trajectory file
    trajectory = ase.io.read(trajectory_file, index=':')  # Read all frames

    # Create a list to store frames with periodic cells
    frames_with_cells = []

    # Loop over each frame in the trajectory
    for atoms in trajectory:
        # Get the positions of all atoms in the structure
        positions = atoms.get_positions()

        # Find the minimum and maximum coordinates in each direction (x, y, z)
        min_pos = np.min(positions, axis=0)
        max_pos = np.max(positions, axis=0)

        # Calculate the size of the box needed to enclose the molecule with padding
        cell_lengths = max_pos - min_pos + 2 * padding

        # Set the periodic cell with the calculated size
        atoms.set_cell(cell_lengths)
        atoms.center()  # Center the atoms inside the cell

        # Enable periodic boundary conditions
        atoms.set_pbc(True)

        # Append the modified frame to the list
        frames_with_cells.append(atoms)

    # Write all frames with periodic cells to a new trajectory file
    ase.io.write(output_file, frames_with_cells)

def main():
    parser = argparse.ArgumentParser(description="A script for adding PBC to each frame of a given trajectory in order to be used as a periodic system (fairchem requires this). Provide the original trajecotry name, a suffix for the new trajectory and the amount of padding.")

    # Mandatory argument
    parser.add_argument("original_traj", type=str, help="Original trajectory to add PBC to")

    # Optional arguments
    parser.add_argument("--newtraj_suffix", type=str, default="_pbc.xyz", help="Resulting filename, by default just adds '_pbc' suffix to the original trajectory name")
    parser.add_argument("--padding", type=float, default=100.00, help="Padding size for the PBC to be added. Use a big number to ensure images don't see each other for non-periodic systems.")
    args = parser.parse_args()

    newtraj = args.original_traj.split(".")[0] + args.newtraj_suffix
    
    # Accessing the arguments
    print("Original traj:", args.original_traj)
    print("Resulting traj:", newtraj)
    print("Size of padding (Ang):", args.padding)

    # Example usage:
    # Replace 'input.traj' with your actual trajectory file, and 'output.traj' with the desired output file
    # We can do it directly with the previous produced splits. 
    create_periodic_cell_for_trajectory(args.original_traj,newtraj, padding=args.padding)

if __name__ == "__main__":
    main()


import argparse
import random
from ase.io import read, write
import os


def split_trajectory(
    file_path, split_ratio=0.8, seed=None, train_output=None, val_output=None
):
    # Set the random seed if provided, for reproducibility
    if seed is not None:
        random.seed(seed)

    # Load the frames from the trajectory file
    frames = read(file_path, index=":")
    num_frames = len(frames)
    print(f"Loaded {num_frames} frames from '{file_path}'.")

    # Shuffle and split frames based on the split ratio
    indices = list(range(num_frames))
    random.shuffle(indices)

    # Calculate the split index
    split_index = int(num_frames * split_ratio)
    train_indices = indices[:split_index]
    val_indices = indices[split_index:]

    # Create train and validation splits
    train_frames = [frames[i] for i in train_indices]
    val_frames = [frames[i] for i in val_indices]

    # Set default output filenames if not provided
    if train_output is None:
        train_output = f"{os.path.splitext(file_path)[0]}_train.xyz"
    if val_output is None:
        val_output = f"{os.path.splitext(file_path)[0]}_val.xyz"

    # Write the split trajectories to separate files
    write(train_output, train_frames)
    write(val_output, val_frames)

    print(
        f"Saved {len(train_frames)} frames to '{train_output}' and {len(val_frames)} frames to '{val_output}'."
    )


def main():
    parser = argparse.ArgumentParser(
        description="Split an ASE trajectory file into train and validation sets in a reproducible way."
    )

    # Mandatory argument
    parser.add_argument(
        "file_path", type=str, help="Path to the input .xyz trajectory file."
    )

    # Optional arguments
    parser.add_argument(
        "--seed",
        type=int,
        default=None,
        help="Random seed for reproducibility (default: None).",
    )
    parser.add_argument(
        "--split_ratio",
        type=float,
        default=0.8,
        help="Ratio of train to validation split (default: 0.8 for 80%% train).",
    )
    parser.add_argument(
        "--train_output",
        type=str,
        default=None,
        help="Output filename for the train set (default: 'input_train.xyz').",
    )
    parser.add_argument(
        "--val_output",
        type=str,
        default=None,
        help="Output filename for the validation set (default: 'input_val.xyz').",
    )

    args = parser.parse_args()

    # Validate the split ratio
    if not 0 < args.split_ratio < 1:
        parser.error("split_ratio must be between 0 and 1 (exclusive).")

    # Call the split function with the provided arguments
    split_trajectory(
        args.file_path, args.split_ratio, args.seed, args.train_output, args.val_output
    )


if __name__ == "__main__":
    main()



import shutil
import tempfile
from unittest.mock import MagicMock, patch

import torch


# Utility function to create a temporary directory
def create_temp_dir() -> str:
    return tempfile.mkdtemp()


# Utility function to clean up a directory
def cleanup_dir(temp_dir: str):
    shutil.rmtree(temp_dir)


def are_dicts_close(dict1, dict2, rtol=1e-4, atol=1e-6, verbose=False):
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return False

    if set(dict1.keys()) != set(dict2.keys()):
        if verbose:
            print(set(dict1.keys()), set(dict2.keys()))
        return False

    for key in dict1.keys():
        if isinstance(dict1[key], dict) and isinstance(dict2[key], dict):
            if not are_dicts_close(dict1[key], dict2[key], rtol, atol):
                return False
        elif isinstance(dict1[key], torch.Tensor) and isinstance(
            dict2[key], torch.Tensor
        ):
            if not torch.allclose(dict1[key], dict2[key], rtol=rtol, atol=atol):
                if verbose:
                    print(f"{key} not equal:\n(1) {dict1[key]}\n(2) {dict2[key]}")
                return False
        else:
            if verbose:
                print("The dictionaries have differnt topology")
            return False
    return True


def mocked_gnn(device, dtype, feature_dim: int = 32, backbone_id: str = "test"):
    # A bunch of code to initialize a mock for the GNN
    gnn = MagicMock()
    gnn.feature_dim = MagicMock(return_value=feature_dim)
    fake_gnn_weight = torch.randn(3, feature_dim, device=device, dtype=dtype)

    def mock_descriptors(data):
        return torch.sin(data.atom_pos) @ fake_gnn_weight

    gnn.descriptors = mock_descriptors

    def load_checkpoint_patch(*args, **kwargs):
        gnn.init_args = MagicMock(return_value=dict(kwargs))
        return gnn

    return patch.multiple("franken.rf.model", load_checkpoint=load_checkpoint_patch)



import numpy as np
import torch

from franken.data.base import Target
from franken.metrics.base import BaseMetric
from franken.metrics.registry import registry


__all__ = ["EnergyMAE", "ForcesMAE", "ForcesCosineSimilarity", "is_pareto_efficient"]


class EnergyMAE(BaseMetric):
    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV",
            "outputs": "meV/atom",
        }
        super().__init__("energy_MAE", device, dtype, units)

    def update(self, predictions: Target, targets: Target) -> None:
        if targets.forces is None:
            raise NotImplementedError(
                "At the moment, target's forces are required to get the number of atoms in the configuration."
            )
        num_atoms = targets.forces.shape[0]
        num_samples = 1
        if targets.energy.ndim > 0:
            num_samples = targets.energy.shape[0]

        error = 1000 * torch.abs(targets.energy - predictions.energy) / num_atoms
        self.buffer_add(error, num_samples=num_samples)


class ForcesMAE(BaseMetric):
    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV/ang",
            "outputs": "meV/ang",
        }
        super().__init__("forces_MAE", device, dtype, units)

    def update(self, predictions: Target, targets: Target) -> None:
        if targets.forces is None or predictions.forces is None:
            raise AttributeError("Forces must be specified to compute the MAE.")
        num_samples = 1
        if targets.forces.ndim > 2:
            num_samples = targets.forces.shape[0]
        elif targets.forces.ndim < 2:
            raise ValueError("Forces must be a 2D tensor or a batch of 2D tensors.")

        error = 1000 * torch.abs(targets.forces - predictions.forces)
        error = error.mean(dim=(-1, -2))  # Average over atoms and components
        self.buffer_add(error, num_samples=num_samples)


class ForcesCosineSimilarity(BaseMetric):
    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
        units = {
            "inputs": "eV/ang",
            "outputs": None,
        }
        super().__init__("forces_cosim", device, dtype, units)

    def update(
        self,
        predictions: Target,
        targets: Target,
    ) -> None:
        num_samples = 1
        assert targets.forces is not None
        assert predictions.forces is not None
        if targets.forces.ndim > 2:
            num_samples = targets.forces.shape[0]
        elif targets.forces.ndim < 2:
            raise ValueError("Forces must be a 2D tensor or a batch of 2D tensors.")

        cos_similarity = torch.nn.functional.cosine_similarity(
            predictions.forces, targets.forces, dim=-1
        )
        cos_similarity = cos_similarity.mean(dim=-1)
        self.buffer_add(cos_similarity, num_samples=num_samples)


def is_pareto_efficient(costs):
    """
    Find the pareto-efficient points
    :param costs: An (n_points, n_costs) array
    :return: A (n_points, ) boolean array, indicating whether each point is Pareto efficient
    """
    is_efficient = np.ones(costs.shape[0], dtype=bool)
    for i, c in enumerate(costs):
        if is_efficient[i]:
            is_efficient[is_efficient] = np.any(
                costs[is_efficient] < c, axis=1
            )  # Keep any point with a lower cost
            is_efficient[i] = True  # And keep self
    return is_efficient


registry.register("energy_MAE", EnergyMAE)
registry.register("forces_MAE", ForcesMAE)
registry.register("forces_cosim", ForcesCosineSimilarity)



import json
import os
from pathlib import Path
import random

import numpy
import pytest
import torch

from franken import FRANKEN_DIR
from franken.backbones.utils import CacheDir, download_checkpoint


__all__ = [
    "ROOT_PATH",
    "DATASETS_PATH",
    "DATASET_REGISTRY",
    "DEFAULT_GNN_IDS",
    "SKIP_NO_CUDA",
    "DEVICES",
    "DEV_CPU_FAIL"
]

ROOT_PATH = FRANKEN_DIR
DATASETS_PATH = ROOT_PATH / "datasets"
with open(DATASETS_PATH / "registry.json", "r") as f:
    DATASET_REGISTRY = json.load(f)

DEFAULT_GNN_IDS = [
    "MACE-L0"
]  # , "SchNet-S2EF-OC20-All"]  # List of gnn_ids to download

SKIP_NO_CUDA = pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")

DEVICES = [
    "cpu",
    pytest.param("cuda:0", marks=SKIP_NO_CUDA) # type: ignore
]
DEV_CPU_FAIL = [
    pytest.param(dev, marks=pytest.mark.xfail(run=False, reason="Not implemented on CPU"))
    if dev == "cpu"
    else dev
    for dev in DEVICES
]

def prepare_gnn_checkpoints():
    # Ensure each gnn_id backbone is downloaded
    for gnn_id in DEFAULT_GNN_IDS:
        download_checkpoint(gnn_id)
    return CacheDir.get() / "gnn_checkpoints"

def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """

    # Cache-dir is either FRANKEN_CACHE_DIR if specified or in the repository folder.
    CacheDir.initialize(
        os.environ.get("FRANKEN_CACHE_DIR", Path(__file__).parent / ".franken")
    )

    prepare_gnn_checkpoints()


@pytest.fixture(autouse=True)
def random_seed():
    """This fixture is called before each test and sets random seeds"""
    random.seed(14)
    numpy.random.seed(14)
    torch.manual_seed(14)



from typing import Final

import torch
from mace.modules.models import MACE
from mace.modules.utils import get_edge_vectors_and_lengths
from e3nn.util.jit import compile_mode

from franken.data import Configuration


@compile_mode("script")
class FrankenMACE(torch.nn.Module):
    interaction_block: Final[int]

    def __init__(self, base_model: MACE, interaction_block, gnn_backbone_id):
        super().__init__()
        self.base_model = base_model
        self.gnn_backbone_id = gnn_backbone_id
        self.interaction_block = interaction_block
        self.interactions = self.base_model.interactions[: self.interaction_block]
        self.products = self.base_model.products[: self.interaction_block]
        self.atomic_numbers = self.base_model.atomic_numbers

    def init_args(self):
        return {
            "gnn_backbone_id": self.gnn_backbone_id,
            "interaction_block": self.interaction_block,
        }

    def descriptors(self, data: Configuration) -> torch.Tensor:
        if self.interaction_block > len(self.base_model.interactions):
            raise ValueError(
                f"This model has {len(self.base_model.interactions)} gnn layers, while descriptors have been required for the {self.interaction_block} layer"
            )
        # assert on local variables to make torchscript happy
        edge_index = data.edge_index
        shifts = data.shifts
        node_attrs = data.node_attrs
        assert edge_index is not None
        assert shifts is not None
        assert node_attrs is not None
        # Embeddings
        node_feats = self.base_model.node_embedding(node_attrs)
        vectors, lengths = get_edge_vectors_and_lengths(
            positions=data.atom_pos,
            edge_index=edge_index,
            shifts=shifts,
        )
        edge_attrs = self.base_model.spherical_harmonics(vectors)
        edge_feats = self.base_model.radial_embedding(
            lengths, node_attrs, edge_index, self.base_model.atomic_numbers
        )

        node_feats_list = []
        for interaction, product in zip(self.interactions, self.products):
            node_feats, sc = interaction(
                node_attrs=node_attrs,
                node_feats=node_feats,
                edge_attrs=edge_attrs,
                edge_feats=edge_feats,
                edge_index=edge_index,
            )

            node_feats = product(node_feats=node_feats, sc=sc, node_attrs=node_attrs)
            # Extract only scalars. Use `irreps_out` attribute to figure out which features correspond to scalars.
            # irreps_out is an `Irreps` object: a 2-tuple of multiplier and `Irrep` objects.
            # Tuple[Tuple[int, Tuple[int, int]], Tuple[int, Tuple[int, int]]]
            # The `Irrep` object is a tuple consisting of parameters `l` and `p`.
            # The scalar irrep is the first in `irreps_out`. Its dimension is computed
            # as `mul * ir.dim` where `ir.dim == 2 * ir.l  + 1`
            # Note this is equivalent code, which does not support TorchScript.
            # invariant_slices = product.linear.irreps_out.slices()[0]
            irreps = product.linear.irreps_out
            invariant_slices = slice(0, irreps[0][0] * (2 * irreps[0][1][0] + 1))
            node_feats_list.append(node_feats[..., invariant_slices])
        return torch.cat(node_feats_list, dim=-1)

    def num_params(self) -> int:
        return sum(p.numel() for p in self.base_model.parameters())

    def feature_dim(self):
        return (
            self.interaction_block
            * self.base_model.node_embedding.linear.irreps_out.count("0e")
        )

    @staticmethod
    def load_from_checkpoint(
        trainer_ckpt, gnn_backbone_id: str, interaction_block: int, map_location=None
    ) -> "FrankenMACE":
        mace = torch.load(
            trainer_ckpt, map_location=map_location, weights_only=False
        ).to(dtype=torch.float32)
        return FrankenMACE(
            base_model=mace,
            gnn_backbone_id=gnn_backbone_id,
            interaction_block=interaction_block,
        )



from collections import namedtuple

import fairchem.core.common.utils
import torch
from fairchem.core.models.schnet import SchNetWrap

from franken.data import Configuration


def segment_coo_patch(src, index, dim_size=None):
    if dim_size is None:
        dim_size = index.max().item() + 1
    out = torch.zeros(dim_size, dtype=src.dtype, device=src.device)
    out.scatter_add_(dim=0, index=index, src=src)
    return out


def segment_csr_patch(src, indptr):
    out = torch.zeros(indptr.size(0) - 1, dtype=src.dtype, device=src.device)
    for i in range(len(indptr) - 1):
        out[i] = src[indptr[i] : indptr[i + 1]].sum()
    return out


def patch_fairchem():
    """
    The `segment_coo` and `segment_csr` patches are necessary to allow
    forward-mode autodiff through the network, which is not implemented
    in the original torch-scatter functions.
    """
    fairchem.core.common.utils.segment_coo = segment_coo_patch
    fairchem.core.common.utils.segment_csr = segment_csr_patch


FairchemCompatData = namedtuple(
    "FairchemCompatData", ["pos", "cell", "batch", "natoms", "atomic_numbers"]
)


class FrankenSchNetWrap(SchNetWrap):
    def __init__(self, *args, interaction_block, gnn_backbone_id, **kwargs):
        patch_fairchem()
        super().__init__(*args, **kwargs)

        self.interaction_block = interaction_block
        self.gnn_backbone_id = gnn_backbone_id

    def descriptors(
        self,
        data: Configuration,
    ):
        """
        Forward pass for the SchNet model to get the embedded representations of the input data
        """
        fairchem_compat_data = FairchemCompatData(
            data.atom_pos, data.cell, data.batch_ids, data.natoms, data.atomic_numbers
        )
        # fairchem checks if the attribute exists, not whether it's None.
        if data.pbc is not None:
            fairchem_compat_data.pbc = data.pbc  # type: ignore
        # Get the atomic numbers of the input data
        z = data.atomic_numbers.long()
        assert z.dim() == 1
        # Get the edge index, edge weight and other attributes of the input data
        graph = self.generate_graph(fairchem_compat_data)

        edge_attr = self.distance_expansion(graph.edge_distance)

        # Get the embedded representations of the input data
        h = self.embedding(z)
        for interaction in self.interactions[: self.interaction_block]:
            h = h + interaction(h, graph.edge_index, graph.edge_distance, edge_attr)

        return h

    def feature_dim(self):
        return self.hidden_channels

    def num_params(self) -> int:
        return sum(p.numel() for p in self.parameters())

    def init_args(self):
        return {
            "gnn_backbone_id": self.gnn_backbone_id,
            "interaction_block": self.interaction_block,
        }

    @staticmethod
    def load_from_checkpoint(
        trainer_ckpt, gnn_backbone_id, interaction_block
    ) -> "FrankenSchNetWrap":
        ckpt_data = torch.load(
            trainer_ckpt, map_location=torch.device("cpu"), weights_only=False
        )

        model_config = ckpt_data["config"]["model_attributes"]
        model_config["otf_graph"] = True

        model = FrankenSchNetWrap(
            **model_config,
            interaction_block=interaction_block,
            gnn_backbone_id=gnn_backbone_id,
        )
        # Before we can load state, need to fix state-dict keys:
        # Match the "module." count in the keys of model and checkpoint state_dict
        # DataParallel model has 1 "module.",  DistributedDataParallel has 2 "module."
        # Not using either of the above two would have no "module."
        ckpt_key_count = next(iter(ckpt_data["state_dict"])).count("module")
        mod_key_count = next(iter(model.state_dict())).count("module")
        key_count_diff = mod_key_count - ckpt_key_count
        if key_count_diff > 0:
            new_dict = {
                key_count_diff * "module." + k: v
                for k, v in ckpt_data["state_dict"].items()
            }
        elif key_count_diff < 0:
            new_dict = {
                k[len("module.") * abs(key_count_diff) :]: v
                for k, v in ckpt_data["state_dict"].items()
            }
        else:
            new_dict = ckpt_data["state_dict"]
        model.load_state_dict(new_dict)
        return model



import pytest
import torch

from franken.rf.heads import (
    BiasedOrthogonalRFF,
    Linear,
    MultiScaleOrthogonalRFF,
    OrthogonalRFF,
    TensorSketch,
)

RF_PARAMETRIZE = [
    "poly",
    "gaussian",
    "linear",
    "biased-gaussian",
    "multiscale-gaussian",
]


def init_rf(rf_type: str, *args, **kwargs):
    if rf_type == "poly":
        return TensorSketch(*args, **kwargs)
    elif rf_type == "gaussian":
        return OrthogonalRFF(*args, **kwargs)
    elif rf_type == "linear":
        return Linear(*args, **kwargs)
    elif rf_type == "biased-gaussian":
        return BiasedOrthogonalRFF(*args, **kwargs)
    elif rf_type == "multiscale-gaussian":
        return MultiScaleOrthogonalRFF(*args, **kwargs)
    else:
        raise ValueError(rf_type)


@pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
class TestDtype:
    @pytest.mark.parametrize("dt", [torch.float32, torch.float64])
    def test_dtype_match(self, dt, rf_type):
        rf = init_rf(
            rf_type,
            input_dim=64,
        )
        data = torch.randn(10, 64, dtype=dt)
        atomic_nums = torch.randint(1, 100, (10,))
        fmap = rf.feature_map(data, atomic_numbers=atomic_nums)
        assert fmap.dtype == dt
        for buf_name, buf in rf.named_buffers():
            if buf_name == "weights":
                # weights not touched by this test so they'll be f32
                assert (
                    buf.dtype == torch.get_default_dtype()
                ), f"weights has unexpected type {buf.dtype}"
            elif buf.numel() > 1 and buf.dtype.is_floating_point:
                assert buf.dtype == dt, f"Buffer {buf_name} has incorrect dtype."


class TestFeatureSizes:
    def test_orff_offset(self):
        rf_offset = init_rf(
            "gaussian", input_dim=32, use_offset=True, num_random_features=128
        )
        rf_no_offset = init_rf(
            "gaussian", input_dim=32, use_offset=False, num_random_features=128
        )
        assert rf_offset.num_random_features == 128
        assert rf_no_offset.num_random_features == 128
        assert rf_offset.total_random_features == 128
        assert rf_no_offset.total_random_features == 256
        assert rf_offset.rff_matrix.shape == (128, 32)
        assert rf_no_offset.rff_matrix.shape == (128, 32)
        assert rf_offset.random_offset.shape == (128,)

    @pytest.mark.parametrize("rf_type", ["poly", "gaussian"])
    def test_per_species_kernel_nonlin1(self, rf_type):
        rf = init_rf(
            rf_type,
            input_dim=32,
            num_random_features=128,
            num_species=4,
            chemically_informed_ratio=None,
        )
        assert rf.num_random_features == 128
        assert rf.total_random_features == 128 * 4

    @pytest.mark.parametrize("rf_type", ["poly", "gaussian"])
    def test_per_species_kernel_nonlin2(self, rf_type):
        rf = init_rf(
            rf_type,
            input_dim=32,
            num_random_features=128,
            num_species=4,
            chemically_informed_ratio=0.4,
        )
        assert rf.num_random_features == 128
        assert rf.total_random_features == 128 * (4 + 1)

    def test_per_species_kernel_lin1(self):
        rf = init_rf(
            "linear", input_dim=32, num_species=4, chemically_informed_ratio=None
        )
        assert rf.num_random_features == 33
        assert rf.total_random_features == 33 * 4

    def test_per_species_kernel_lin2(self):
        rf = init_rf(
            "linear", input_dim=32, num_species=4, chemically_informed_ratio=0.4
        )
        assert rf.num_random_features == 33
        assert rf.total_random_features == (33) * (4 + 1)


class TestEdgeCaseInputs:
    @pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
    def test_zero_lengthscale(self, rf_type):
        with pytest.raises(ValueError):
            if rf_type != "multiscale-gaussian":
                init_rf(rf_type, input_dim=32, length_scale=0)
            else:
                init_rf(rf_type, input_dim=32, length_scale_low=0)

    @pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
    def test_negative_lengthscale(self, rf_type):
        with pytest.raises(ValueError):
            if rf_type != "multiscale-gaussian":
                init_rf(rf_type, input_dim=32, length_scale=-1.1)
            else:
                init_rf(rf_type, input_dim=32, length_scale_low=-1.1)


if __name__ == "__main__":
    pytest.main()



"""Utilities for dealing with triangular matrices in-place."""

import math
from typing import Optional, Tuple, Union

import numpy as np
import torch

try:
    import triton.language as tl
    from triton import jit
except ImportError:

    def jit(*a, **kw):
        pass

    class mock_tl:
        constexpr = None

    tl = mock_tl()


def naive_lerp(
    start: torch.Tensor,
    end: torch.Tensor,
    weight: Union[float, torch.Tensor],
    inplace: bool = False,
) -> torch.Tensor:
    r"""Linear interpolation wrapper.
    Computes the linear interpolation between :attr:`start` and :attr:`end` based on
    the specified :attr:`weight`. If :attr:`inplace` is True, the result is stored overwriting
    the :attr:`start` tensor.

    .. math::
        \text{out}_i = \text{start}_i + \text{weight}_i \times (\text{end}_i - \text{start}_i)

    Args:
        input (Tensor): the tensor with the starting points
        end (Tensor): the tensor with the ending points
        weight (float or tensor): the weight for the interpolation formula
        inplace (bool): whether the interpolation should be computed in place.
    """
    if not inplace:
        return torch.lerp(start, end, weight)
    return start.lerp_(end, weight)


def triangular_lerp(
    matrix: torch.Tensor,
    diag_upper: torch.Tensor,
    diag_lower: torch.Tensor,
    weight: float,
    inplace: bool = False,
) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""Linearly interpolate the upper and lower triangular parts of a matrix.
    The diagonals of the upper and lower triangles must be specified separately using
    the :attr:`diag_upper` and :attr:`diag_lower` arguments. If :attr:`inplace` is True,
    the result is stored overwriting :attr:`diag_upper` and the upper triangle
    of :attr:`matrix`.

    The first output is the linear interpolation of the triangular matrix

    .. math:
        \text{out}_1 = \text{upper}(\text{matrix}) + \text{weight} \times (\text{lower}(\text{matrix}) - \text{upper}(\text{matrix}))

    the second output is the linear interpolation of the diagonals

    .. math:
        \text{out}_2 = \text{diag_upper} + \text{weight} \times (\text{diag_lower} - \text{diag_upper})

    Args:
        matrix (Tensor): the square matrix whose triangular parts are interpolated
        diag_upper (Tensor): a vector containing the start point of interpolation
        diag_lower (Tensor): a vector containing the end point of interpolation
        weight (float or tensor): the weight for the interpolation formula
        inplace (bool): whether the interpolation should be computed in place.
    """
    diag_lerped = naive_lerp(diag_upper, diag_lower, weight, inplace)
    if matrix.device.type == "cpu":
        triang_lerped = _trilerp_cpu(
            matrix, weight=weight, out=matrix if inplace else None
        )
    else:
        triang_lerped = _trilerp_triton(
            matrix, weight=weight, out=matrix if inplace else None
        )
    return triang_lerped, diag_lerped


def _trilerp_cpu(
    matrix: torch.Tensor, weight: float, out: Optional[torch.Tensor] = None
) -> torch.Tensor:
    up_ids = np.triu_indices(matrix.shape[0], k=1)

    if out is None:
        out = torch.empty_like(matrix)
    out_np = out.numpy()
    matrix_np = matrix.numpy()
    lerped = matrix_np[up_ids] + weight * (matrix_np.T[up_ids] - matrix_np[up_ids])
    out_np[up_ids] = lerped
    return torch.from_numpy(out_np)


@jit
def trilerp_kernel(
    a_ptr,
    N,
    stride_a1,
    stride_a2,
    out_ptr,
    stride_o1,
    stride_o2,
    weight,
    BLOCK_SIZE_N: tl.constexpr,
):
    # TODO: Check if this could be replaced by cublas GEAM (a cuda specific extension!). This is in cupy!
    # Compute the block index in the upper triangle that we're
    # going to compute
    pid = tl.program_id(axis=0)
    # triangular program ID to block ID
    row = ((tl.sqrt_rn((8 * pid + 1).to(tl.float32)) - 1.0) / 2.0).to(tl.int32)
    pid_a1 = pid - row * (row + 1) // 2
    pid_a2 = row

    offs_a1_up = pid_a1 * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_a2_up = pid_a2 * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    a_up_ptrs = a_ptr + (
        offs_a1_up[:, None] * stride_a1 + offs_a2_up[None, :] * stride_a2
    )
    offs_a1_lo = offs_a2_up
    offs_a2_lo = offs_a1_up
    a_lo_ptrs = a_ptr + (
        offs_a1_lo[:, None] * stride_a1 + offs_a2_lo[None, :] * stride_a2
    )
    out_ptrs = out_ptr + (
        offs_a1_up[:, None] * stride_o1 + offs_a2_up[None, :] * stride_o2
    )

    mask = (
        (offs_a1_up[:, None] < offs_a2_up[None, :])
        & (offs_a1_up[:, None] < N)
        & (offs_a2_up[None, :] < N)
    )
    up = tl.load(a_up_ptrs, mask=mask, other=0.0)
    lo = tl.load(a_lo_ptrs, mask=mask.T, other=0.0)
    interp = up * (1 - weight) + lo.T * weight
    tl.store(out_ptrs, interp, mask=mask)


def _trilerp_triton(
    matrix: torch.Tensor, weight: float, out: Optional[torch.Tensor] = None
) -> torch.Tensor:
    assert matrix.shape[0] == matrix.shape[1], "Incompatible dimensions"
    N = matrix.shape[0]
    s1 = matrix.stride(0)
    s2 = matrix.stride(1)
    if out is None:
        out = torch.empty_like(matrix)
    o1 = out.stride(0)
    o2 = out.stride(1)
    block_size = 64  # fixed, this is not a hot-loop function.
    grid = (math.ceil(N / block_size) * (math.ceil(N / block_size) + 1) // 2,)
    trilerp_kernel[grid](matrix, N, s1, s2, out, o1, o2, weight, block_size)  # type: ignore
    return out


@jit
def inplace_triangular_divide_kernel(
    a_ptr, N, stride_a1, stride_a2, val, upper, BLOCK_SIZE_N: tl.constexpr
):
    pid = tl.program_id(axis=0)
    # triangular program ID to block ID
    row = ((tl.sqrt_rn((8 * pid + 1).to(tl.float32)) - 1.0) / 2.0).to(tl.int32)
    pid_a1 = pid - row * (row + 1) // 2
    pid_a2 = row
    if upper:
        offs_a1_up = pid_a1 * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
        offs_a2_up = pid_a2 * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
        mask = (
            (offs_a1_up[:, None] < offs_a2_up[None, :])
            & (offs_a1_up[:, None] < N)
            & (offs_a2_up[None, :] < N)
        )
        a_ptrs = a_ptr + (
            offs_a1_up[:, None] * stride_a1 + offs_a2_up[None, :] * stride_a2
        )
    else:
        offs_a1_lo = pid_a2 * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
        offs_a2_lo = pid_a1 * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
        mask = (
            (offs_a1_lo[:, None] > offs_a2_lo[None, :])
            & (offs_a1_lo[:, None] < N)
            & (offs_a2_lo[None, :] < N)
        )
        a_ptrs = a_ptr + (
            offs_a1_lo[:, None] * stride_a1 + offs_a2_lo[None, :] * stride_a2
        )
    tri_mat = tl.load(a_ptrs, mask=mask, other=0.0)
    scaled_tri_mat = tri_mat / val
    tl.store(a_ptrs, scaled_tri_mat, mask=mask)


def _cpu_triangular_divide(
    matrix: torch.Tensor, value: float, upper: bool
) -> torch.Tensor:
    if upper:
        tri_ids = np.triu_indices(matrix.shape[0], k=1)
    else:
        tri_ids = np.tril_indices(matrix.shape[0], k=-1)
    matrix_np = matrix.numpy()
    matrix_np[tri_ids] /= value
    return torch.from_numpy(matrix_np)


def inplace_triangular_divide(
    matrix: torch.Tensor, value: float, upper: bool
) -> torch.Tensor:
    """Divide a triangular matrix by a scalar inplace.

    The diagonal of the matrix is not touched by this function.

    Args:
        matrix (Tensor): a square matrix
        value (float): the divisor
        upper (bool): whether to divide the upper or lower triangle of :attr:`matrix`.

    """
    assert matrix.shape[0] == matrix.shape[1], "Incompatible dimensions"
    if matrix.device.type == "cpu":
        matrix = _cpu_triangular_divide(matrix, value, upper)
    else:
        N = matrix.shape[0]
        s1 = matrix.stride(0)
        s2 = matrix.stride(1)
        block_size = 64
        grid = (math.ceil(N / block_size) * (math.ceil(N / block_size) + 1) // 2,)
        inplace_triangular_divide_kernel[grid](
            matrix, N, s1, s2, value, upper, block_size
        )  # type: ignore
    return matrix



import json
import logging
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Union

import numpy as np
import torch

import franken.metrics
import franken.utils.distributed as dist_utils


logger = logging.getLogger("franken")


class DataSplit(Enum):
    UNDEFINED = -1
    TRAIN = 0
    TEST = 1
    VALIDATION = 2
    VAL = 2  # alias


@dataclass(frozen=True)
class MetricLog:
    split: DataSplit
    name: str
    value: float


@dataclass(frozen=True)
class HyperParameter:
    name: str
    value: Union[float, bool, str, int]


@dataclass(frozen=True)
class HyperParameterGroup:
    group_name: str
    hyperparameters: list[HyperParameter]

    @classmethod
    def from_dict(cls, group_name: str, hp_dict: dict):
        params = []
        for param_name, param_value in hp_dict.items():
            params.append(HyperParameter(name=param_name, value=param_value))
        return cls(group_name=group_name, hyperparameters=params)


@dataclass
class LogEntry:
    checkpoint_hash: str
    checkpoint_rf_weight_id: int
    timings_cov_coeffs: float
    timings_solve: float
    metrics: list[MetricLog] = field(default_factory=list)
    hyperparameters: list[HyperParameterGroup] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict) -> "LogEntry":
        """Create a LogEntry instance from a dict."""

        # Parse checkpoint
        checkpoint_hash = data["checkpoint"]["hash"]
        checkpoint_rf_weight_id = data["checkpoint"]["rf_weight_id"]

        # Parse timings
        timings_cov_coeffs = data["timings"]["cov_coeffs"]
        timings_solve = data["timings"]["solve"]

        # Parse metrics
        metrics = []
        for split_name, split_metrics in data["metrics"].items():
            try:
                split = DataSplit[split_name.upper()]
            except KeyError:
                split = DataSplit.UNDEFINED

            for metric_name, value in split_metrics.items():
                metrics.append(MetricLog(split=split, name=metric_name, value=value))

        # Parse hyperparameters
        hyperparameters = []
        for group_name, group_params in data["hyperparameters"].items():
            hp_group = HyperParameterGroup.from_dict(group_name, group_params)
            hyperparameters.append(hp_group)

        return cls(
            checkpoint_hash=checkpoint_hash,
            checkpoint_rf_weight_id=checkpoint_rf_weight_id,
            timings_cov_coeffs=timings_cov_coeffs,
            timings_solve=timings_solve,
            metrics=metrics,
            hyperparameters=hyperparameters,
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert this LogEntry instance to dict."""

        # Convert metrics to the expected format
        metrics_dict = defaultdict(dict)
        for metric in self.metrics:
            metrics_dict[metric.split.name.lower()][metric.name] = metric.value

        # Convert hyperparameters to the expected format
        hyperparams_dict = {}
        for group in self.hyperparameters:
            hyperparams_dict[group.group_name] = {
                hp.name: hp.value for hp in group.hyperparameters
            }

        # Create the final dictionary structure
        data = {
            "checkpoint": {
                "hash": self.checkpoint_hash,
                "rf_weight_id": self.checkpoint_rf_weight_id,
            },
            "timings": {
                "cov_coeffs": self.timings_cov_coeffs,
                "solve": self.timings_solve,
            },
            "metrics": metrics_dict,
            "hyperparameters": hyperparams_dict,
        }

        return data

    def get_metric(self, name: str, split: Union[DataSplit, str] = DataSplit.TRAIN):
        if isinstance(split, str):
            try:
                split = DataSplit[split.upper()]
            except KeyError:
                raise KeyError(f"Unknown split {split=}")
        matches = []
        for metric in self.metrics:
            if metric.name == name and metric.split == split:
                matches.append(metric)
        if len(matches) == 0:
            raise KeyError(f"Unable to find the metric {name=} on {split=}")
        elif len(matches) == 1:
            return matches[0].value
        else:
            raise NameError(f"Multiple metrics with {name=} and {split=} found.")

    def get_hyperparameter(self, name: str, group: str):
        matches = []
        for hp_group in self.hyperparameters:
            for hp in hp_group.hyperparameters:
                if hp_group.group_name == group and hp.name == name:
                    matches.append(hp)
        if len(matches) == 0:
            raise KeyError(f"Unable to find the hyperparameter {group=}/{name=}")
        elif len(matches) == 1:
            return matches[0].value
        else:
            raise NameError(
                f"Multiple hyperparameters with {group=} and {name=} found."
            )

    def add_metric(self, name: str, value: float, split: Union[DataSplit, str]):
        if isinstance(split, str):
            try:
                split = DataSplit[split.upper()]
            except KeyError:
                raise KeyError(f"Unknown split {split=}")
        try:
            self.get_metric(name, split=split)
            raise ValueError(
                f"A metric with the same {name=} and {split=} already exists"
            )
        except KeyError:
            pass
        self.metrics.append(MetricLog(split=split, name=name, value=value))

    def __eq__(self, other) -> bool:
        try:
            same_hash = self.checkpoint_hash == other.checkpoint_hash
            same_id = self.checkpoint_rf_weight_id == other.checkpoint_rf_weight_id
            return same_hash and same_id
        except AttributeError:
            return False

    def __str__(self):
        return (
            f"<LogEntry hash={self.checkpoint_hash}, id={self.checkpoint_rf_weight_id}>"
        )


@dataclass
class LogCollection:
    logs: list[LogEntry]

    def __len__(self):
        return len(self.logs)

    def __iter__(self):
        return iter(self.logs)

    def __getitem__(self, idx):
        return self.logs[idx]

    def get_metric(
        self,
        name: str,
        split: Union[DataSplit, str] = DataSplit.TRAIN,
        default=float("nan"),
    ):
        values = []
        for entry_id, entry in enumerate(self.logs):
            try:
                metric_value = entry.get_metric(name, split=split)
            except KeyError:
                logger.warning(
                    f"Metric {name} not found for the {entry_id+1}-th log, returning {default}"
                )
                metric_value = default
            values.append(metric_value)
        return values

    def get_hyperparameter(
        self,
        name: str,
        group: str,
        default=float("nan"),
    ):
        values = []
        for entry_id, entry in enumerate(self.logs):
            try:
                hp_value = entry.get_hyperparameter(name, group)
            except KeyError:
                logger.warning(
                    f"The hyperparameter {group}/{name} was not found for the {entry_id+1}-th log, returning {default}"
                )
                hp_value = default
            values.append(hp_value)
        return values

    @classmethod
    def gather_from_ranks(cls, local_logs: dict[int, LogEntry]):
        idx_list = []
        entries_list = []
        for local_log in dist_utils.all_gather_object(local_logs):
            for idx, log_entry in local_log.items():
                idx_list.append(idx)
                entries_list.append(log_entry)
        assert sorted(idx_list) == list(
            range(len(idx_list))
        ), "The indices of the local logs fail to form a continuous list from 0 to sum(len(logs) for log in local_logs)"
        logger.debug(f"Retrieved {len(idx_list)} logs.")

        sorted_entries = [None for _ in idx_list]
        for idx, entry in zip(idx_list, entries_list):
            sorted_entries[idx] = entry
        return cls(logs=sorted_entries)

    def save_json(self, log_file: Path) -> None:
        """Append the log entry to a JSON file and save the full logs file"""
        log_batch = [log.to_dict() for log in self.logs]
        if log_file.exists():
            with open(log_file, "r+") as f:
                logs = json.load(f)
                logs.extend(log_batch)
                f.seek(0)
                json.dump(logs, f, indent=4, cls=dtypeJSONEncoder)
        else:
            with open(log_file, "w") as f:
                json.dump(log_batch, f, indent=4, cls=dtypeJSONEncoder)

    @classmethod
    def from_json(cls, log_file: Path):
        with open(log_file, "r") as f:
            logs = json.load(f)
        if not isinstance(logs, list):  # handle loading a single log!
            logs = [logs]
        return cls(logs=[LogEntry.from_dict(log) for log in logs])

    def get_best_model(
        self,
        metrics_to_minimize: list[str] = ["energy_MAE", "forces_MAE"],
        p: int = 1,
        split: DataSplit = DataSplit.TRAIN,
    ) -> LogEntry:
        """
        Find the best model from a list of log entries.
        The best model is chosen among the pareto frontier of ``metrics_to_minimize``,
        by minimizing their ``p``-norm.
        The function returns a dictionary with information about the best model.
        """
        available_metrics = franken.metrics.available_metrics()
        for metric in metrics_to_minimize:
            assert metric in available_metrics, f"Unknown {metric=}"
        costs = np.stack(
            [self.get_metric(m, split=split) for m in metrics_to_minimize], axis=-1
        )
        costs = np.nan_to_num(costs, nan=np.inf, posinf=np.inf, neginf=-np.inf)
        err_norm = np.linalg.norm(costs, ord=p, axis=-1)
        err_norm[~franken.metrics.is_pareto_efficient(costs)] = np.inf
        best_model_idx = np.argmin(err_norm)
        return self.logs[best_model_idx]


class dtypeJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, torch.dtype):
            return str(o)
        return super().default(o)



import warnings
from typing import Optional, Tuple

import scipy.linalg.blas
import torch


try:
    import cupy
    import cupy.cublas
except ImportError:
    cupy = None

from franken.utils.linalg.tri import inplace_triangular_divide


def rank1_update(
    cov: torch.Tensor, diag: torch.Tensor, vec: torch.Tensor, upper: bool = True
) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""in-place rank-1 update to a triangular matrix

    If cupy is available, a low-memory-footprint algorithm is used.

    Returns the updated covariance and diagonal vector:

    .. math::
        \text{out}_1 = \text{cov} + \text{vec} \text{vec}^T

    .. math::
        \text{out}_2 = \text{diag} + \text{vec}^2

    Args:
        cov (Tensor): the square matrix which is updated in-place
        diag (Tensor): the diagonal of the triangular matrix
        vec (Tensor): the vector comprising the symmetric rank-1 update
        upper (bool): whether to update the lower or upper triangle of :attr:`cov`.
    """
    if cupy is not None or cov.device.type == "cpu":
        return _lowmemcov_rank1_update(cov, diag, vec, upper)
    warnings.warn(
        "low-memory covariance updates cannot be used "
        "because `cupy` is not available. Install `cupy` "
        "if you encounter memory problems."
    )
    return _rank1_update(cov, diag, vec, upper)


def _rank1_update(
    cov: torch.Tensor, diag: torch.Tensor, vec: torch.Tensor, upper: bool = True
) -> Tuple[torch.Tensor, torch.Tensor]:
    tri_fn = torch.triu if upper else torch.tril
    cov.add_(tri_fn(torch.outer(vec, vec)))
    diag.add_(vec**2)
    return cov, diag


def _lowmemcov_rank1_update(
    cov: torch.Tensor, diag: torch.Tensor, vec: torch.Tensor, upper: bool = True
) -> Tuple[torch.Tensor, torch.Tensor]:
    # Use the DSYR function in blas
    # A := alpha*x*x**T + A
    assert vec.device == cov.device
    assert vec.dtype == cov.dtype
    if cov.device.type == "cpu":
        # cov needs to be F-contiguous for the overwriting to work
        transposed = False
        if cov.shape[0] > 1 and cov.stride(1) == 1:  # then it's C-contiguous
            cov = cov.T
            upper = not upper
            transposed = True
        # Note that cov_np and cov share the same memory, so changes in cov_np are reflected in cov.
        cov_np = cov.numpy()
        energy_fmap_np = vec.numpy()
        if cov.dtype == torch.float64:
            scipy.linalg.blas.dsyr(
                alpha=1.0, x=energy_fmap_np, a=cov_np, lower=not upper, overwrite_a=True
            )
        elif cov.dtype == torch.float32:
            scipy.linalg.blas.ssyr(
                alpha=1.0, x=energy_fmap_np, a=cov_np, lower=not upper, overwrite_a=True
            )
        else:
            raise ValueError(cov.dtype)
        if transposed:
            cov = cov.T
    else:
        assert cupy is not None
        energy_fmap_cp = cupy.asarray(vec.view(-1, 1))
        cov_cp = cupy.asarray(cov)
        # It's a bit dumb but cupy doesn't have great coverage of the cuBLAS level 2 API, so
        # we use 'syrk' again instead of 'syr'.
        cupy.cublas.syrk(
            trans="N",
            a=energy_fmap_cp,
            out=cov_cp,
            alpha=1.0,
            beta=1.0,
            lower=not upper,
        )
    diag.add_(vec**2)
    return cov, diag


def rankk_update(
    cov: torch.Tensor, diag: torch.Tensor, update: torch.Tensor, upper: bool = True
) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""in-place rank-k update to a triangular matrix.

    If cupy is available, a low-memory-footprint algorithm is used.

    Returns the updated covariance and diagonal vector:

    .. math::
        \text{out}^{(1)} = \text{cov} + \text{update} \text{update}^T

    .. math::
        \text{out}^{(2)}_i = \text{diag}_i + \sum_{j} \text{update}_{i,j}^2

    Args:
        cov (Tensor): the square matrix which is updated in-place
        diag (Tensor): the diagonal of the triangular matrix
        update (Tensor): the rank-k update matrix
        upper (bool): whether to update the lower or upper triangle of :attr:`cov`.
    """
    if cupy is not None or cov.device.type == "cpu":
        return _lowmemcov_rankk_update(cov, diag, update, upper)
    warnings.warn(
        "low-memory covariance updates cannot be used "
        "because `cupy` is not available. Install `cupy` "
        "if you encounter memory problems."
    )
    return _rankk_update(cov, diag, update, upper)


def _rankk_update(
    cov: torch.Tensor, diag: torch.Tensor, update: torch.Tensor, upper: bool = True
):
    tri_fn = torch.triu if upper else torch.tril
    cov.add_(tri_fn(update @ update.T))
    diag.add_((update**2).sum(1))
    return cov, diag


def _lowmemcov_rankk_update(
    cov: torch.Tensor, diag: torch.Tensor, update: torch.Tensor, upper: bool = True
):
    # DSYRK
    # C := alpha*A*A**T + beta*C,
    assert update.device == cov.device
    assert update.dtype == cov.dtype

    if cov.device.type == "cpu":
        # cov needs to be F-contiguous for the overwriting to work
        transposed = False
        if cov.shape[0] > 1 and cov.stride(1) == 1:  # then it's C-contiguous
            cov = cov.T
            upper = not upper
            transposed = True
        # Note that cov_np and cov share the same memory, so changes in cov_np are reflected in cov.
        cov_np = cov.numpy()
        forces_fmap_np = update.numpy()
        if cov.dtype == torch.float64:
            scipy.linalg.blas.dsyrk(
                alpha=1.0,
                a=forces_fmap_np,
                beta=1.0,
                c=cov_np,
                lower=not upper,
                overwrite_c=True,
            )
        elif cov.dtype == torch.float32:
            scipy.linalg.blas.ssyrk(
                alpha=1.0,
                a=forces_fmap_np,
                beta=1.0,
                c=cov_np,
                lower=not upper,
                overwrite_c=True,
            )
        else:
            raise ValueError(cov.dtype)
        if transposed:
            cov = cov.T
    else:
        assert cupy is not None
        forces_fmap_cp = cupy.asarray(update)
        cov_cp = cupy.asarray(cov)
        cupy.cublas.syrk(
            trans="N",
            a=forces_fmap_cp,
            out=cov_cp,
            alpha=1.0,
            beta=1.0,
            lower=not upper,
        )

    diag.add_((update**2).sum(1))
    return cov, diag


def triu_to_cov(triu: torch.Tensor, cov: Optional[torch.Tensor] = None) -> torch.Tensor:
    tril_indices = torch.tril_indices(triu.shape[0], triu.shape[1])
    if cov is None:
        cov = triu.clone()
    else:
        assert cov.shape == triu.shape
        cov.copy_(triu)
    cov[tril_indices.unbind()] = cov.T[tril_indices.unbind()]
    return cov


def tril_to_cov(tril: torch.Tensor, cov: Optional[torch.Tensor] = None) -> torch.Tensor:
    tril_indices = torch.tril_indices(tril.shape[0], tril.shape[1])
    if cov is None:
        cov = tril.clone()
    else:
        assert cov.shape == tril.shape
        cov.copy_(tril)
    cov.T[tril_indices.unbind()] = cov[tril_indices.unbind()]
    return cov


def normalize_leading_eig(cov: torch.Tensor, coeffs: torch.Tensor) -> None:
    r"""Normalize covariance matrix dividing by the largest eigenvalue in-place"""
    norm, _ = torch.lobpcg(cov, k=1, largest=True)
    cov.div_(norm)
    coeffs.div_(norm)


def lowmem_normalize_leading_eig(
    tri_cov: torch.Tensor, diag: torch.Tensor, coeffs: torch.Tensor, upper: bool
) -> None:
    r"""Normalize low-memory covariance matrix dividing by the largest eigenvalue.

    The low-memory covariance matrix is stored only in one triangle of :attr:`tri_cov`.

    We need to compute the eigenvalues from the triangular parts of the matrix.
    Since this is not really possible with LOBPCG, we convert to full matrices
    and compute the eigenvalues of those.
    TODO: This requires a new copy of `cov` which wastes a bunch of memory.
    """
    tri_cov.diagonal().copy_(diag)
    if upper:
        cov = triu_to_cov(tri_cov)
    else:
        cov = tril_to_cov(tri_cov)
    norm, _ = torch.lobpcg(cov, k=1, largest=True)
    diag.div_(norm[0])
    coeffs.div_(norm[0])
    inplace_triangular_divide(tri_cov, norm[0].item(), upper=upper)



import hashlib
import logging
from pathlib import Path

import hydra
from omegaconf import DictConfig
import torch

from franken.autotune.script import init_loaders
from franken.backbones.utils import CacheDir
from franken.rf.model import FrankenPotential
from franken.trainers.log_utils import LogCollection
from franken.trainers.rf_cuda_lowmem import RandomFeaturesTrainer
import franken.utils.distributed as dist_utils
from franken.utils.misc import pprint_config, setup_logger


logger = logging.getLogger("franken")


@hydra.main(version_base="1.3", config_path="configs", config_name="evaluate.yaml")
def _evaluate(cfg: DictConfig):
    """
    Load a model and evaluate it on a given dataset
    """
    # Load all needed configuration keys immediately
    logging_level = cfg.get("console_logging_level", "WARNING").upper()
    data_path = Path(cfg.data_path)
    model_path = Path(cfg.model_path)
    model_info_path = (
        Path(cfg.model_info_path) if cfg.get("model_info_path") is not None else None
    )
    rf_weight_id = cfg.get("rf_weight_id", 0)
    log_out_path = Path(cfg.save_path)
    run_dir = Path(cfg.run_dir) if cfg.get("run_dir") is not None else None
    if run_dir is None:
        run_dir = model_path if model_path.is_dir() else model_path.parent

    if torch.cuda.is_available():
        rank = dist_utils.init()
        device = torch.device(torch.cuda.current_device())
    else:
        rank = 0
        device = torch.device("cpu")
    # Create run folder (from a single rank)
    if rank != 0:
        run_dir = None
        dist_utils.barrier()
    else:
        run_dir.mkdir(parents=True, exist_ok=True)
        dist_utils.barrier()  # other ranks follow

    setup_logger(
        level=logging_level,
        directory=dist_utils.broadcast_obj(run_dir),
        rank=rank,
        logname="franken_eval",
    )
    pprint_config(cfg)

    if log_out_path.is_dir():
        raise ValueError(
            f"'save_path' '{log_out_path.resolve()}' is a directory. It must be a file."
        )
    if not log_out_path.name.endswith(".json"):
        log_out_path = Path(str(log_out_path.resolve()) + ".json")
    if log_out_path.exists():
        logger.warning(
            f"'save_path' '{log_out_path.resolve()}' already exists and will be overwritten."
        )

    CacheDir.initialize()

    # Figure out the path to the model
    if model_path.is_dir():
        _model_path = model_path / "best_ckpt.pt"
        if not _model_path.exists():
            raise FileNotFoundError(
                f"Inferred model path '{_model_path.resolve()}' does not exist."
            )
    else:
        _model_path = model_path
        if not _model_path.exists():
            raise FileNotFoundError(
                f"Model path '{_model_path.resolve()}' does not exist."
            )

    # Load model. Splitting the ranks here since initializing FrankenPotential
    # can result in backbone download. We avoid double download.
    if rank != 0:
        dist_utils.barrier()
    logger.info(f"Loading FrankenPotential from '{_model_path.resolve()}'")
    franken = FrankenPotential.load(
        _model_path,
        map_location=device,
        rf_weight_id=rf_weight_id,
    )
    model_hash = hashlib.md5(str(franken.hyperparameters).encode())
    model_hash = model_hash.hexdigest()
    if rank == 0:
        dist_utils.barrier()

    data_loader = init_loaders(
        franken.gnn_backbone_id,
        train_path=None,
        test_path=data_path,
    )["test"]

    # Figure out path to the model info (LogEntry corresponding to the model)
    if model_info_path is None:
        if model_path.is_dir() and (model_path / "best.json").exists():
            model_info_path = model_path / "best.json"
    elif model_info_path.is_dir():
        model_info_path = model_info_path / "log.json"
    if model_info_path is None:
        raise FileNotFoundError(
            "Could not infer model_info_path. Pass it to the program explicitly."
        )
    if not model_info_path.exists():
        raise FileNotFoundError(
            f"Could not find explicit model_info_path at '{model_info_path.resolve()}'"
        )

    # Load logs and check correspondence to model
    logger.info(f"Loading LogCollection from '{model_info_path.resolve()}'")
    log_collection = LogCollection.from_json(model_info_path)
    model_info = None
    for log_entry in log_collection:
        if log_entry.checkpoint_hash == model_hash:
            model_info = log_entry
            break
    if model_info is None and len(log_collection) == 1:
        logger.warning("Model hash does not correspond with log hash")
        model_info = log_collection[0]
    elif model_info is None:
        raise ValueError(
            f"Could not find model-info at '{model_info_path.resolve()}' corresponding to loaded franken model"
        )

    trainer = RandomFeaturesTrainer(
        train_dataloader=None,  # type: ignore
        log_dir=None,
        device=device,
    )
    eval_logs = trainer.evaluate(
        model=franken,
        dataloader=data_loader,
        log_collection=LogCollection([model_info]),
        all_weights=None,
    )
    if rank == 0:
        eval_logs.save_json(log_out_path)
        logger.info(f"Saved evaluation logs to {log_out_path}")


def evaluate():
    _evaluate()


if __name__ == "__main__":
    evaluate()



import importlib.resources
import json
import logging
import os
from pathlib import Path

import requests
import torch
import torch.utils.data
from tqdm import tqdm


logger = logging.getLogger("franken")


def load_model_registry():
    model_registry_text = (
        importlib.resources.files("franken.backbones")
        .joinpath("registry.json")
        .read_text()
    )
    model_registry = json.loads(model_registry_text)
    return model_registry


class CacheDir:
    directory: Path | None = None

    @staticmethod
    def initialize(cache_dir: Path | str | None = None):
        if CacheDir.is_initialized():
            logger.warning(
                f"Cache directory already initialized at {CacheDir.directory}. Reinitializing."
            )
        # Default cache location: ~/.franken
        default_cache = Path.home() / ".franken"
        if cache_dir is None:
            env_cache_dir = os.environ.get("FRANKEN_CACHE_DIR", None)
            if env_cache_dir is None:
                logger.info(f"Initializing default cache directory at {default_cache}")
                cache_dir = default_cache
            else:
                logger.info(
                    f"Initializing cache directory from $FRANKEN_CACHE_DIR {env_cache_dir}"
                )
                cache_dir = env_cache_dir
        else:
            logger.info(f"Initializing custom cache directory {cache_dir}")
        CacheDir.directory = Path(cache_dir)

        # Ensure the directory exists
        if not CacheDir.directory.exists():
            CacheDir.directory.mkdir(parents=True, exist_ok=True)
            logger.info(f"Created cache directory at: {CacheDir.directory}")

    @staticmethod
    def get() -> Path:
        if not CacheDir.is_initialized():
            CacheDir.initialize()
        assert CacheDir.directory is not None
        return CacheDir.directory

    @staticmethod
    def is_initialized() -> bool:
        return CacheDir.directory is not None


def make_summary(cache_dir: str | None = None):
    """Function to print available models, first those present locally."""
    if cache_dir is not None:
        CacheDir.initialize(cache_dir=cache_dir)
    registry = load_model_registry()
    ckpt_dir = CacheDir.get() / "gnn_checkpoints"

    local_models = []
    remote_models = []
    _summary = ""
    for model, info in registry.items():
        local_path = ckpt_dir / info["local"]
        kind = info["kind"]
        implemented = info.get("implemented", False)
        if implemented:
            if local_path.is_file():
                local_models.append((model, kind))
            else:
                remote_models.append((model, kind))
    if len(local_models) > 0:
        _summary += f"{'DOWNLOADED MODELS':^80}\n"
        _summary += f"{'(' + str(ckpt_dir) + ')':-^80}\n"
        for model, kind in local_models:
            _str = f"{model} ({kind})"
            _summary += f"{_str:<40}\n"

    _summary += f"{'AVAILABLE MODELS':-^80}\n"
    for model, kind in remote_models:
        _str = f"{model} ({kind})"
        _summary += f"{_str:<80}\n"
    _summary += "-" * 80
    return _summary


def get_checkpoint_path(gnn_backbone_id: str) -> Path:
    registry = load_model_registry()
    gnn_checkpoints_dir = CacheDir.get() / "gnn_checkpoints"

    if gnn_backbone_id not in registry.keys():
        raise NameError(
            f"Unknown {gnn_backbone_id} GNN backbone, the current available backbones are\n{make_summary()}"
        )
    else:
        backbone_info = registry[gnn_backbone_id]
        ckpt_path = gnn_checkpoints_dir / backbone_info["local"]
        if not ckpt_path.exists():
            download_checkpoint(gnn_backbone_id)
    return ckpt_path


def download_checkpoint(gnn_backbone_id: str, cache_dir: str | None = None) -> None:
    """Download the model if it's not already present locally."""
    registry = load_model_registry()
    if cache_dir is not None:
        CacheDir.initialize(cache_dir=cache_dir)
    ckpt_dir = CacheDir.get() / "gnn_checkpoints"

    if gnn_backbone_id not in registry.keys():
        raise NameError(
            f"Unknown {gnn_backbone_id} GNN backbone, the current available backbones are\n{make_summary()}"
        )

    if not registry[gnn_backbone_id]["implemented"]:
        raise NotImplementedError(
            f"The model {gnn_backbone_id} is not implemented in franken yet."
        )

    local_path = ckpt_dir / registry[gnn_backbone_id]["local"]
    remote_path = registry[gnn_backbone_id]["remote"]

    if local_path.is_file():
        logger.info(
            f"Model already exists locally at {local_path}. No download needed."
        )
        return

    local_path.parent.mkdir(parents=True, exist_ok=True)

    logger.info(f"Downloading model from {remote_path} to {local_path}")
    try:
        response = requests.get(remote_path, stream=True)
        response.raise_for_status()

        total_size = int(response.headers.get("content-length", 0))

        with (
            open(local_path, "wb") as f,
            tqdm(
                desc="Downloading",
                total=total_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                ncols=100,
            ) as bar,
        ):
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
                bar.update(len(chunk))
    except requests.RequestException as e:
        logger.error(f"Download failed. {e}")
        raise e


def load_checkpoint(
    gnn_backbone_id: str,
    **kwargs,
) -> torch.nn.Module:
    ckpt_path = get_checkpoint_path(gnn_backbone_id)
    registry = load_model_registry()

    backbone_family = registry[gnn_backbone_id]["kind"].lower()
    err_msg = f"franken wasn't able to load {gnn_backbone_id}. Is {backbone_family} installed?"
    if backbone_family == "fairchem":
        try:
            from franken.backbones.wrappers.fairchem_schnet import FrankenSchNetWrap
        except ImportError as import_err:
            logger.error(err_msg, exc_info=import_err)
            raise
        return FrankenSchNetWrap.load_from_checkpoint(
            str(ckpt_path), gnn_backbone_id=gnn_backbone_id, **kwargs
        )
    elif backbone_family == "mace":
        try:
            from franken.backbones.wrappers.mace_wrap import FrankenMACE
        except ImportError as import_err:
            logger.error(err_msg, exc_info=import_err)
            raise
        return FrankenMACE.load_from_checkpoint(
            str(ckpt_path), gnn_backbone_id=gnn_backbone_id, **kwargs
        )
    elif backbone_family == "sevenn":
        try:
            from franken.backbones.wrappers.sevenn import FrankenSevenn
        except ImportError as import_err:
            logger.error(err_msg, exc_info=import_err)
            raise
        return FrankenSevenn.load_from_checkpoint(
            ckpt_path, gnn_backbone_id=gnn_backbone_id, **kwargs
        )
    else:
        raise ValueError(f"Unknown backbone family {backbone_family}")



import logging

import torch


logger = logging.getLogger("franken")


def patch_e3nn():
    # NOTE:
    #  Patching should occur when doing training: it is necessary for `jvp` on the MACE model,
    #  but not during inference, when we only use `torch.autograd`. For inference, we may want
    #  to compile the model using `torch.jit` - and the patch interferes with the JIT, so we
    #  must disable it.

    # Ugly patch code. Need to replace a scripted function, but it's messy.
    import e3nn.o3._spherical_harmonics

    new_locals = {"Tensor": torch.Tensor}
    if not hasattr(e3nn.o3._spherical_harmonics._spherical_harmonics, "code"):
        logger.warning(
            "Patching e3n '_spherical_harmonics' function failed unexpectedly. "
            "This may or may not be a problem."
        )
        return
    exec(e3nn.o3._spherical_harmonics._spherical_harmonics.code, None, new_locals)

    def _spherical_harmonics(
        lmax: int, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor
    ) -> torch.Tensor:
        return new_locals["_spherical_harmonics"](torch.tensor(lmax), x, y, z)

    e3nn.o3._spherical_harmonics._spherical_harmonics = _spherical_harmonics



import logging
from unittest import mock

import ase.calculators.lj
import numpy as np
from omegaconf import DictConfig
import pytest
from ase.build import molecule


from franken.mdgen.asemd import run_md_simulation



@pytest.fixture()
def atoms():
    return molecule("H2O")


class NaNCalculator(ase.calculators.lj.LennardJones):
    def calculate(self, atoms=None, properties=None, system_changes=...):
        super().calculate(atoms, properties, system_changes)
        self.results["forces"] = np.full_like(self.results["forces"], np.nan)


def test_nan_catcher(atoms, tmp_path, caplog):
    calc = NaNCalculator()
    atoms.calc = calc

    with caplog.at_level(logging.ERROR):
        with mock.patch.object(calc, "calculate", wraps=calc.calculate) as calc_mock:
            run_md_simulation(
                output_dir=tmp_path,
                integrator_cfg=DictConfig({
                    "_target_": "ase.md.langevin.Langevin",
                    "friction": 0.1,
                    "timestep": 0.01,
                    "temperature_K": 300,
                }),
                initial_velocity_cfg=DictConfig({
                    "_target_": "ase.md.velocitydistribution.MaxwellBoltzmannDistribution",
                    "temperature_K": 300
                }),
                initial_configuration=atoms,
                num_steps=20,
                save_every=1,
                log_every=1,
                seed=4,
            )
            calc_mock.assert_called_once()
    assert "Simulation failed after" in caplog.text



from ase.io import read, write
from numpy.random import shuffle, seed


if __name__ == "__main__":
    traj = read("dataset.traj", index=":")

    seed(42)
    shuffle(traj)

    train = traj[:30000]
    valid = traj[30000:31000]
    test = traj[31000:]

    write("train.extxyz", train)
    write("valid.extxyz", valid)
    write("test.extxyz", test)



import pytest

from franken.trainers.log_utils import HyperParameterGroup, LogEntry


@pytest.fixture
def dummy_log_dict():
    return {
        "checkpoint": {"hash": "rand_uuid", "rf_weight_id": 0},
        "timings": {"cov_coeffs": 1.0, "solve": 1.0},
        "metrics": {
            "train": {"energy_MAE": 1.0, "forces_MAE": 1.0, "forces_cosim": 1.0},
            "validation": {"energy_MAE": 1.0, "forces_MAE": 1.0, "forces_cosim": 1.0},
            "test": {"energy_MAE": 1.0, "forces_MAE": 1.0, "forces_cosim": 1.0},
        },
        "hyperparameters": {
            "franken": {
                "gnn_backbone_id": "SchNet-S2EF-OC20-All",
                "interaction_block": 3,
                "kernel_type": "gaussian",
            },
            "random_features": {
                "num_random_features": 1024,
            },
            "input_scaler": {"scale_by_Z": True, "num_species": 2},
            "solver": {
                "L2_penalty": 1e-6,
                "loss_lerp_weight": 0.1,
                "dtype": "torch.float64",
            },
        },
    }


def test_hpgroup_from_dict():
    dummy_group_dict = {
        "str_param": "str_value",
        "int_param": 1,
        "float_param": 1.0,
        "bool_param": True,
    }

    hpg = HyperParameterGroup.from_dict("dummy_group", dummy_group_dict)
    assert hpg.group_name == "dummy_group"
    for hp in hpg.hyperparameters:
        assert hp.name in dummy_group_dict.keys()
        assert hp.value == dummy_group_dict[hp.name]


def test_log_entry_serialize_deserialize(dummy_log_dict):
    log_entry = LogEntry.from_dict(dummy_log_dict)
    assert log_entry.to_dict() == dummy_log_dict


def test_log_entry_get_metric(dummy_log_dict):
    log_entry = LogEntry.from_dict(dummy_log_dict)
    assert log_entry.get_metric("energy_MAE", "train") == 1.0


def test_log_entry_get_invalid_metric_name(dummy_log_dict):
    log_entry = LogEntry.from_dict(dummy_log_dict)
    with pytest.raises(KeyError):
        log_entry.get_metric("invalid_metric", "train")


def test_log_entry_get_invalid_metric_split(dummy_log_dict):
    log_entry = LogEntry.from_dict(dummy_log_dict)
    with pytest.raises(KeyError):
        log_entry.get_metric("energy_MAE", "invalid_split")


# class TestBestModel:
#     def test_all_nans(self):
#         log_entries = [
#             {"metrics": {"val": {"energy": torch.nan}}},
#             {"metrics": {"val": {"energy": torch.nan}}},
#         ]
#         expected_best_log = log_entries[0]
#         best_log = get_best_model(log_entries, ["energy"], split="val")
#         assert best_log == expected_best_log

#     def test_nans(self):
#         log_entries = [
#             {"metrics": {"val": {"energy": torch.nan}}},
#             {"metrics": {"val": {"energy": 0.1}}},
#             {"metrics": {"val": {"energy": 12.0}}},
#         ]
#         expected_best_log = log_entries[1]
#         best_log = get_best_model(log_entries, ["energy"], split="val")
#         assert best_log == expected_best_log
#         log_entries = [
#             {"metrics": {"val": {"energy": 0.1}}},
#             {"metrics": {"val": {"energy": torch.nan}}},
#             {"metrics": {"val": {"energy": 12.0}}},
#         ]
#         expected_best_log = log_entries[0]
#         best_log = get_best_model(log_entries, ["energy"], split="val")
#         assert best_log == expected_best_log

#     def test_stability(self):
#         log_entries = [
#             {"metrics": {"val": {"energy": 1.0, "forces": 12}}},
#             {"metrics": {"val": {"energy": 1.1, "forces": 11.9}}},
#             {"metrics": {"val": {"energy": 1.2, "forces": 11.8}}},
#         ]
#         expected_best_log = log_entries[0]
#         best_log = get_best_model(log_entries, ["energy", "forces"], split="val")
#         assert best_log == expected_best_log

#     def test_normal(self):
#         log_entries = [
#             {"metrics": {"val": {"energy": 1.0, "forces": 12}}},
#             {"metrics": {"val": {"energy": 0.9, "forces": 11.9}}},
#             {"metrics": {"val": {"energy": 1.2, "forces": 11.8}}},
#         ]
#         expected_best_log = log_entries[1]
#         best_log = get_best_model(log_entries, ["energy", "forces"], split="val")
#         assert best_log == expected_best_log

#     def test_missing_split(self):
#         log_entries = [
#             {"metrics": {"val": {"energy": 1.0, "forces": 12}}},
#         ]
#         with pytest.raises(KeyError):
#             get_best_model(log_entries, ["energy", "forces"], split="train")

#     def test_missing_metric(self):
#         log_entries = [
#             {"metrics": {"val": {"energy": 1.0, "forces": 12}}},
#         ]
#         with pytest.raises(KeyError):
#             get_best_model(log_entries, ["missing", "forces"], split="val")



import os
import pytest
import torch.distributed
from torch.multiprocessing import Process, Pipe, SimpleQueue

from franken.data.base import Configuration
from franken.data.fairchem import FairchemAtomsDataset
from .conftest import (
    DATASET_REGISTRY,
    DATASETS_PATH,
)

class ThrowingProcess(Process):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pconn, self._cconn = Pipe()
        self._exception = None

    def run(self):
        try:
            super().run()
            self._cconn.send(None)
        except Exception as e:
            self._cconn.send(e)
            raise e

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception


def init_processes(rank, size, fn, backend='gloo'):
    """ Initialize the distributed environment. """
    os.environ['MASTER_ADDR'] = '127.0.0.64'
    os.environ['MASTER_PORT'] = '26512'
    os.environ['GLOO_SOCKET_IFNAME'] = "lo"
    torch.distributed.init_process_group(backend, rank=rank, world_size=size)
    fn()


def init_distributed_cpu(num_proc, run_fn):
    processes = []
    for rank in range(num_proc):
        p = ThrowingProcess(target=init_processes, args=(rank, num_proc, run_fn))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()
        if p.exception:
            error = p.exception
            raise error


def mocked_dataset(num_atoms, dtype, device, num_configs: int = 1):
    data = []
    for _ in range(num_configs):
        data.append(Configuration(
            torch.randn(num_atoms, 3, dtype=dtype),
            torch.randint(1, 100, (num_atoms,)),
            torch.tensor(num_atoms),
        ).to(device))
    return data


@pytest.mark.parametrize("num_samples", [1, 7, 19])
@pytest.mark.parametrize("num_procs", [1, 4])
def test_distributed_dataloader_length(num_samples, num_procs):
    def inner_fn():
        data_path = DATASETS_PATH / DATASET_REGISTRY["test"]["long"]
        dataset = FairchemAtomsDataset(
            data_path,
            split="train",
            num_random_subsamples=num_samples,
            subsample_rng=None,
            gnn_backbone_id=None,
            precompute=False,  # no point running a2g
        )
        assert len(dataset) == num_samples
        dataloader = dataset.get_dataloader(True)
        rank = torch.distributed.get_rank()
        ws = torch.distributed.get_world_size()
        assert len(dataloader) == (len(dataset) // ws) + int(len(dataset) % ws > rank)

    init_distributed_cpu(num_procs, inner_fn)


def test_distributed_dataloader_order():
    num_samples = 7
    num_procs = 3
    ids_queue = SimpleQueue()
    def inner_fn():
        data_path = DATASETS_PATH / DATASET_REGISTRY["test"]["long"]
        dataset = FairchemAtomsDataset(
            data_path,
            split="train",
            num_random_subsamples=num_samples,
            subsample_rng=None,
            gnn_backbone_id=None,
            precompute=True,
        )
        assert len(dataset) == num_samples
        dataloader = dataset.get_dataloader(True)
        rank = torch.distributed.get_rank()
        dl_elements = [el for el in dataloader]
        dl_id = 0
        for i in range(rank, num_samples, num_procs):
            torch.testing.assert_close(
                dl_elements[dl_id][0].atom_pos, dataset[i][0].atom_pos
            )
            torch.testing.assert_close(
                dl_elements[dl_id][1].forces, dataset[i][1].forces
            )
            dl_id += 1
            ids_queue.put(i)
        assert dl_id == len(dl_elements)
    init_distributed_cpu(num_procs, inner_fn)
    # Assert all IDs were processed - only once
    all_ids = []
    while not ids_queue.empty():
        all_ids.append(ids_queue.get())
    assert sorted(all_ids) == list(range(num_samples))



import os
from typing import List, Tuple, Union
import warnings
from socket import gethostname

import hostlist
import torch
import torch.distributed


def slurm_to_env():
    hostname = hostlist.expand_hostlist(os.environ["SLURM_JOB_NODELIST"])[0]
    os.environ["MASTER_ADDR"] = hostname
    os.environ["MASTER_PORT"] = os.environ.get("MASTER_PORT", "33633")
    try:
        world_size = int(os.environ["SLURM_NTASKS"])
    except KeyError:
        world_size = int(os.environ["SLURM_NTASKS_PER_NODE"]) * int(
            os.environ["SLURM_NNODES"]
        )
    os.environ["WORLD_SIZE"] = str(world_size)
    os.environ["LOCAL_RANK"] = os.environ["SLURM_LOCALID"]
    os.environ["RANK"] = os.environ["SLURM_PROCID"]


def is_torchrun():
    # torchrun command sets the env variables RANK, LOCAL_RANK, and WORLD_SIZE
    return "RANK" in os.environ and "WORLD_SIZE" in os.environ


def is_slurm():
    return "SLURM_PROCID" in os.environ


def init(distributed: bool) -> int:
    if distributed:
        if not torch.cuda.is_available():
            raise RuntimeError("Distributed training is only supported on CUDA")
        if is_torchrun():
            pass
        elif is_slurm():
            slurm_to_env()
        else:
            warnings.warn(
                "Cannot initialize distributed training. "
                "Neither torchrun nor SLURM environment variable were found."
            )
        world_size = int(os.environ.get("WORLD_SIZE", 1))
        if world_size > 1:
            print(
                f"Distributed initialization at rank {os.environ['RANK']} of {world_size} "
                f"(rank {os.environ['LOCAL_RANK']} on {gethostname()} with "
                f"{torch.cuda.device_count()} GPUs allocated)."
            )
            torch.distributed.init_process_group(
                backend="nccl",
                device_id=torch.device(f"cuda:{int(os.environ['LOCAL_RANK'])}"),
            )

    device = f"cuda:{get_local_rank()}"
    torch.cuda.set_device(device)
    return get_rank()


def get_local_rank() -> int:
    if torch.distributed.is_initialized():
        return int(os.environ["LOCAL_RANK"])
    return 0


def get_rank() -> int:
    if torch.distributed.is_initialized():
        return torch.distributed.get_rank()
    return 0


def barrier() -> None:
    if torch.distributed.is_initialized():
        torch.distributed.barrier()


def get_world_size() -> int:
    if torch.distributed.is_initialized():
        return torch.distributed.get_world_size()
    return 1


def all_reduce(tensor: torch.Tensor, op) -> None:
    if torch.distributed.is_initialized():
        torch.distributed.all_reduce(tensor, op)
    return None


def all_sum(tensor: torch.Tensor) -> None:
    if torch.distributed.is_initialized():
        torch.distributed.all_reduce(tensor, torch.distributed.ReduceOp.SUM)
    return None


def broadcast_obj(obj, src=0):
    if torch.distributed.is_initialized():
        to_broadcast = [obj]
        torch.distributed.broadcast_object_list(to_broadcast, src=src)
        return to_broadcast[0]
    return obj


def all_gather_into_tensor(
    out_size: Union[Tuple, torch.Size], in_tensor: torch.Tensor
) -> torch.Tensor:
    if torch.distributed.is_initialized():
        out_tensor = torch.zeros(
            out_size, dtype=in_tensor.dtype, device=in_tensor.device
        )
        torch.distributed.all_gather_into_tensor(out_tensor, in_tensor)
        return out_tensor
    return in_tensor


def all_gather(tensor: torch.Tensor) -> List[torch.Tensor]:
    if torch.distributed.is_initialized():
        shapes = [
            tensor.shape if r == get_rank() else None for r in range(get_world_size())
        ]
        for r in range(get_world_size()):
            shapes[r] = broadcast_obj(shapes[r], src=r)
        tensor_list = [
            (
                tensor
                if r == get_rank()
                else torch.empty(shapes[r], device=tensor.device, dtype=tensor.dtype)  # type: ignore
            )  # type: ignore
            for r in range(get_world_size())
        ]
        torch.distributed.all_gather(tensor_list, tensor)
        return tensor_list
    return [tensor]


def all_gather_object(obj) -> List:
    if torch.distributed.is_initialized():
        output = [None for _ in range(get_world_size())]

        torch.distributed.all_gather_object(output, obj)
        return output
    return [obj]


def print0(*args, **kwargs):
    if get_rank() == 0:
        print(*args, **kwargs)



from typing import Optional, Iterator

from torch.utils.data import Sampler, Dataset
import torch.distributed as dist


class SimpleUnevenDistributedSampler(Sampler):
    def __init__(
        self,
        dataset: Dataset,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
    ) -> None:
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        if rank >= num_replicas or rank < 0:
            raise ValueError(
                f"Invalid rank {rank}, rank should be in the interval [0, {num_replicas - 1}]"
            )
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.total_size = len(self.dataset)  # type: ignore[arg-type]
        # num_samples indicates the number of samples for the current process
        self.num_samples = len(range(self.rank, self.total_size, self.num_replicas))

    def __iter__(self) -> Iterator:
        indices = list(range(len(self.dataset)))  # type: ignore[arg-type]

        # subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self) -> int:
        return self.num_samples



import gc
import inspect
import logging
import os
import sys
from itertools import product
from pathlib import Path
from time import perf_counter
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import torch
from omegaconf import DictConfig, ListConfig, OmegaConf

import franken.utils.distributed as dist_utils


logger = logging.getLogger("franken")


def sanitize_init_dict(cls, params: dict) -> dict:
    signature = inspect.signature(cls.__init__)

    # Get valid parameter names, excluding 'self'
    valid_params = set(signature.parameters.keys()) - {"self"}

    # Filter the input params to only include valid parameters
    filtered_params = {
        key: value for key, value in params.items() if key in valid_params
    }

    invalid_params = {
        key: value for key, value in params.items() if key not in valid_params
    }
    if len(invalid_params) > 0:
        logger.warning(
            f"Tried to initialize {cls.__name__} with invalid parameters: {invalid_params}"
        )
    # Initialize and return the class instance
    return filtered_params


def params_grid(
    grid: Mapping[str, Sequence[Any]],
    split_distributed: bool = False,
    filter_function: Optional[
        Callable[[Dict[str, Any]], Optional[Dict[str, Any]]]
    ] = None,
) -> Generator[Tuple[int, Dict[str, Any]], None, None]:
    """Converts a map of hyperparameter names to their possible values to a list of individual hyperparameter combinations

    Args:
        grid (Mapping): maps hyperparameter names to all the values they can take
            in a hyperparameter sweep.
        split_distributed (bool, optional): The result is distributed among available
            processes (uses the pytorch rank and world-size for this). Defaults to False.
        filter_function (callable, optional): A function which takes each set of hyperparameters
            and returns the same hyperparameters, a modified set or None if they are invalid.

    Example::

        >>> hp_grid = {"hp1": ["val1", "val2"], "hp2": [1, 5, 10], "hp3": ["fixed"]}
        >>> list(params_grid(hp_grid))
        [
            (0, {'hp1': 'val1', 'hp2': 1, 'hp3': 'fixed'}),
            (1, {'hp1': 'val1', 'hp2': 5, 'hp3': 'fixed'}),
            (2, {'hp1': 'val1', 'hp2': 10, 'hp3': 'fixed'}),
            (3, {'hp1': 'val2', 'hp2': 1, 'hp3': 'fixed'}),
            (4, {'hp1': 'val2', 'hp2': 5, 'hp3': 'fixed'}),
            (5, {'hp1': 'val2', 'hp2': 10, 'hp3': 'fixed'})
        ]
    """
    # Always sort the keys of a dictionary, for reproducibility
    items = sorted(grid.items())
    if not items:
        return
    else:
        keys, values = zip(*items)
        for idx, v in enumerate(product(*values)):
            params = dict(zip(keys, v))
            if filter_function is not None:
                params = filter_function(params)
            if params is None:
                continue
            if split_distributed:
                if idx % dist_utils.get_world_size() == dist_utils.get_rank():
                    params = dict(zip(keys, v))
                    yield idx, params
                else:
                    continue
            else:
                yield idx, params


class throughput:
    def __init__(
        self,
        iterable,
        desc="",
        units="cfgs",
        update_interval=0.2,
        world_size: int | None = None,  # overrides world-size
        process_rank: int | None = None,  # overrides rank
        total: int | None = None,
        device=None,
        leave: bool = False,
    ):
        self.iterable = iterable
        self.update_interval = update_interval
        self.world_size = (
            dist_utils.get_world_size() if world_size is None else world_size
        )
        self.process_rank = (
            dist_utils.get_rank() if process_rank is None else process_rank
        )
        self.total = len(iterable) if hasattr(iterable, "__len__") else None
        # Initialize time-counters here to avoid errors with None later on.
        self.now = perf_counter()
        self.last_update = perf_counter()
        self.n_iterations = 0
        self._batch = 0
        self.desc = desc
        self.units = units
        self.hardware = get_device_name(device)
        self.total = total
        self.leave = leave

    def __iter__(self):
        if self.process_rank != 0:
            for obj in self.iterable:
                yield obj
            return

        self.last_update = perf_counter()

        try:
            for obj in self.iterable:
                yield obj
                # Update and possibly print the progressbar.
                # Note: does not call self.update(1) for speed optimisation.
                self._batch += 1
                if self.should_print():
                    self.display()
        finally:
            self.display()
            if self.leave:
                print()
            else:
                print("\r", end="")

    def should_print(self):
        self.now = perf_counter()
        return self.now - self.last_update > self.update_interval

    def display(self):
        time_elapsed = self.now - self.last_update
        num_done = self.world_size * self._batch
        throughput = num_done / (time_elapsed + 1e-9)
        self.n_iterations += num_done
        if self.total is not None:
            assert self.total > 0
            perc = 100 * self.n_iterations / self.total
            status = f"\r{self.desc:^20} | {throughput:.1f} {self.units}/s | {perc:>3.0f}% | {self.world_size} x {self.hardware}"
        else:
            status = f"\r{self.desc:^20} | {throughput:.1f} {self.units}/s | {self.world_size} x {self.hardware}"
        print(
            status,
            end="",
        )
        self._batch = 0
        self.last_update = self.now

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def get_device_name(device: str | None):
    if device is None:
        return ""
    if torch.device(device).type == "cpu":
        import os
        import platform
        import re
        import subprocess

        def get_processor_name():
            if platform.system() == "Windows":
                return platform.processor()
            elif platform.system() == "Darwin":
                os.environ["PATH"] = os.environ["PATH"] + os.pathsep + "/usr/sbin"
                command = "sysctl -n machdep.cpu.brand_string"
                return (
                    subprocess.check_output(command, shell=True)
                    .strip()
                    .decode(sys.stdout.encoding)
                )
            elif platform.system() == "Linux":
                command = "cat /proc/cpuinfo"
                all_info = subprocess.check_output(command, shell=True).decode().strip()
                for line in all_info.split("\n"):
                    if "model name" in line:
                        return re.sub(".*model name.*:", "", line, 1)
            return ""

        hardware = get_processor_name()
    elif torch.device(device).type == "cuda":
        hardware = torch.cuda.get_device_name(0)
    else:
        raise ValueError(f"Unrecognized device {device}")
    return hardware


# based on https://github.com/BlackHC/toma/blob/master/toma/torch_cuda_memory.py
def is_cuda_out_of_memory(exception: BaseException) -> bool:
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "CUDA" in exception.args[0]
        and "out of memory" in exception.args[0]
    )


# based on https://github.com/BlackHC/toma/blob/master/toma/torch_cuda_memory.py
def garbage_collection_cuda() -> None:
    """Garbage collection Torch (CUDA) memory."""
    gc.collect()
    try:
        # This is the last thing that should cause an OOM error, but seemingly it can.
        torch.cuda.empty_cache()
    except RuntimeError as exception:
        if not is_cuda_out_of_memory(exception):
            # Only handle OOM errors
            raise


class DDPRankFilter(logging.Filter):
    """Logging filter by rank. Pass `extra={"rank0": True}` when logging"""

    def filter(self, record: logging.LogRecord) -> bool:
        # noinspection PyUnresolvedReferences
        only_rank_0 = hasattr(record, "rank0") and record.rank0 is True

        if dist_utils.get_rank() != 0 and only_rank_0:
            return False
        return True


def remove_root_logger_handlers():
    # This is necessary since some libraries (MACE) configure root loggers
    # which causes double outputs! We remove handlers which are directly
    # attached to the root logger. This is behind a flag, set to true
    # when called from our scripts but potentially user-configurable
    root_logger = logging.getLogger()
    root_logger.handlers.clear()


def setup_logger(
    level: Union[int, str] = logging.INFO,
    directory: Optional[str | Path] = None,
    logname: str = "franken",
    rank: Optional[int] = 0,
):
    # From https://github.com/ACEsuit/mace/blob/main/mace/tools/utils.py
    flogger = logging.getLogger("franken")
    flogger.setLevel(logging.DEBUG)  # Set to DEBUG to capture all levels

    # Add rank-based filtering
    flogger.addFilter(DDPRankFilter())

    # Create formatters
    formatter = logging.Formatter(
        f"%(asctime)s.%(msecs)03d %(levelname)s (rank {rank}): %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # Create console handler
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(level)
    ch.setFormatter(formatter)
    flogger.addHandler(ch)

    if directory is not None:
        directory = str(directory)
        os.makedirs(name=directory, exist_ok=True)

        # Create file handler for non-debug logs
        main_log_path = os.path.join(directory, f"{logname}.log")
        fh_main = logging.FileHandler(main_log_path)
        fh_main.setLevel(logging.DEBUG)  # Dump everything into the logfile
        fh_main.setFormatter(formatter)
        flogger.addHandler(fh_main)
    logging.captureWarnings(True)


def get_logger_stdout_level(log: logging.Logger):
    handlers = [
        h
        for h in log.handlers
        if isinstance(h, logging.StreamHandler) and h.stream == sys.stdout
    ]
    levels = [h.level for h in handlers]
    if len(levels) == 0:
        return 0
    min_level = min(levels)
    return min_level


def pprint_config(cfg: DictConfig | dict):
    tl_keys = sorted(list(cfg.keys()))
    tl_leaf_keys = [
        k
        for k in tl_keys
        if not isinstance(cfg[k], (DictConfig, dict, list, ListConfig))
    ]
    tl_struct_keys = [k for k in tl_keys if k not in tl_leaf_keys]
    s = ""
    for k in tl_leaf_keys:
        s += f"{k}: {str(cfg[k])}\n"
    yaml_content: dict[str, str] = {}
    for key in tl_struct_keys:
        yaml_content[key] = OmegaConf.to_yaml(cfg[key], resolve=True, sort_keys=True)
    for key, content in yaml_content.items():
        s += f"{key}:\n"
        aligned_content = [f"    {line}" for line in content.splitlines(True)]
        s += f"{''.join(aligned_content)}"
    s = s.replace("\n\n", "\n")
    # Output
    logger.debug(s, extra={"rank0": True})
    # If logger has output the string to stdout then we don't print it
    # otherwise we also print it!
    if get_logger_stdout_level(logger) > logging.DEBUG and dist_utils.get_rank() == 0:
        print(s)


def are_lists_equal(l1: Sequence, l2: Sequence) -> bool:
    if len(l1) != len(l2):
        return False
    for i1, i2 in zip(l1, l2):
        if isinstance(i1, dict):
            if not isinstance(i2, dict):
                return False
            if not are_dicts_equal(i1, i2):
                return False
        elif isinstance(i1, (list, tuple)):
            if not isinstance(i2, (list, tuple)):
                return False
            if not are_lists_equal(i1, i2):
                return False
        else:
            if i1 != i2:
                return False
    return True


def are_dicts_equal(d1: dict, d2: dict) -> bool:
    if set(d1.keys()) != set(d2.keys()):
        return False
    for k1, v1 in d1.items():
        if isinstance(v1, dict):
            if not isinstance(d2[k1], dict):
                return False
            if not are_dicts_equal(v1, d2[k1]):
                return False
        elif isinstance(v1, (list, tuple)):
            if not isinstance(d2[k1], (list, tuple)):
                return False
            if not are_lists_equal(v1, d2[k1]):
                return False
        else:
            if v1 != d2[k1]:
                return False
    return True



import logging
import os

os.environ["OMP_NUM_THREADS"] = "8"
from unittest.mock import DEFAULT, patch

import ase
import numpy as np
import pytest
import torch
from ase import units
from ase.io import read
from ase.md.npt import NPT
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution

from franken.calculators.ase_calc import FrankenCalculator
from franken.autotune.script import init_loaders, run_autotune
from franken.data import BaseAtomsDataset
from franken.data.base import Configuration
from franken.rf.model import FrankenPotential
from franken.rf.scaler import Statistics
from franken.trainers.rf_cuda_lowmem import RandomFeaturesTrainer
from franken.utils.misc import garbage_collection_cuda

from .conftest import DATASET_REGISTRY, DATASETS_PATH, DEFAULT_GNN_IDS, DEVICES
from .utils import are_dicts_close, cleanup_dir, create_temp_dir, mocked_gnn

RF_PARAMETRIZE = [
    "poly",
    "gaussian",
    "linear",
    "biased-gaussian",
    "multiscale-gaussian",
]


@pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
@pytest.mark.parametrize("device", DEVICES)
def test_deterministic_initialization(rf_type, device):
    rfparams = {"rng_seed": 42} if rf_type != "linear" else {}
    for gnn_id in DEFAULT_GNN_IDS:
        # Instantiate two models with the same rng_seed
        model1 = FrankenPotential(
            gnn_id,
            rf_type,
            random_features_params=rfparams,
        ).to(device)

        model2 = FrankenPotential(
            gnn_id,
            rf_type,
            random_features_params=rfparams,
        ).to(device)

        # Compare their rf.state_dict()
        assert are_dicts_close(
            model1.rf.state_dict(), model2.rf.state_dict()
        ), f"Model initializations are not deterministic for {device=}, {gnn_id=}, {rf_type=}"


def mocked_torch_save_load():
    return patch.multiple("torch", load=DEFAULT, save=DEFAULT)


def mocked_dataset(num_atoms, dtype, device, num_configs: int = 1):
    data = []
    for _ in range(num_configs):
        data.append(
            Configuration(
                torch.randn(num_atoms, 3, dtype=dtype),
                torch.randint(1, 100, (num_atoms,)),
                torch.tensor(num_atoms),
            ).to(device)
        )
    return data


def test_load_with_legacy_gnn_arg(caplog):
    # Tests whether loading a model where gnn is specified at the base-level
    # of the checkpoint instead of within 'gnn' key
    with mocked_gnn("cpu", torch.float32, backbone_id="test"):
        model = FrankenPotential(
            gnn_backbone_id="test",
            kernel_type="gaussian"
        ).to("cpu")
        save_path = "test.pt"
        with mocked_torch_save_load() as p2:
            model.save(save_path)
            save_ckpt = p2["save"].call_args.args[0]
            assert save_ckpt is not None
            assert save_path == p2["save"].call_args.args[1]
            # Simulate the old model
            save_ckpt['gnn_backbone_id'] = 'test'
            save_ckpt['interaction_block'] = save_ckpt['gnn']['config']['interaction_block']
            del save_ckpt['gnn']
            # Now load and check
            p2["load"].return_value = save_ckpt
            with caplog.at_level(logging.WARNING):
                loaded_model = FrankenPotential.load(save_path)
            assert "Loading legacy checkpoint with limited" in caplog.text
            assert loaded_model.gnn.init_args() == {
                "gnn_backbone_id": "test",
                "interaction_block": 2,
            }


def test_load_with_rf_dtype_arg(caplog):
    # Tests whether loading a model where rf has dtype parameter specified still works.
    # Since the dtype parameter was removed this should just give a warning.
    with mocked_gnn("cpu", torch.float32, backbone_id="test"):
        model = FrankenPotential(
            gnn_backbone_id="test",
            kernel_type="gaussian"
        ).to("cpu")
        save_path = "test.pt"
        with mocked_torch_save_load() as p2:
            model.save(save_path)
            save_ckpt = p2["save"].call_args.args[0]
            assert save_ckpt is not None
            assert save_path == p2["save"].call_args.args[1]
            # Set the dtype argument to simulate an old model
            save_ckpt["rf"]["config"]["dtype"] = torch.float64
            # Now load and check that it only gives a warning
            p2["load"].return_value = save_ckpt
            # NOTE: caplog is a default fixture which allows to capture
            #       the outputs of the logging module.
            with caplog.at_level(logging.WARNING):
                loaded_model = FrankenPotential.load(save_path)
            assert "invalid parameters: {'dtype'" in caplog.text
            assert "dtype" not in loaded_model.rf.init_args()


@pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
@pytest.mark.parametrize("device", DEVICES)
@pytest.mark.parametrize("scale_by_Z", [True, False])
def test_save_load_functionality(rf_type, device, scale_by_Z):
    """Test for checking save and load methods of FrankenPotential"""
    for gnn_id in DEFAULT_GNN_IDS:
        temp_dir = None
        try:
            # Step 1: Create a temporary directory for saving the model
            temp_dir = create_temp_dir()

            data_path = DATASETS_PATH / DATASET_REGISTRY["test"]["test"]
            dataset = BaseAtomsDataset.from_path(
                data_path=data_path,
                split="train",
                gnn_backbone_id=gnn_id,
            )

            model = FrankenPotential(
                gnn_backbone_id=gnn_id,
                kernel_type=rf_type,
                random_features_params={},
                scale_by_Z=scale_by_Z,
                num_species=dataset.num_species,
            ).to(device)

            with torch.no_grad():
                gnn_features_stats = Statistics()
                for data, _ in dataset:  # type: ignore
                    data = data.to(device=device)
                    gnn_features = model.gnn.descriptors(data)
                    gnn_features_stats.update(
                        gnn_features, atomic_numbers=data.atomic_numbers
                    )

                model.input_scaler.set_from_statistics(gnn_features_stats)
                garbage_collection_cuda()

            # Step 2: Save the model to the temporary directory
            model_save_path = os.path.join(temp_dir, "model_checkpoint.pth")
            model.save(model_save_path)

            # Step 3: Load the model from the saved checkpoint
            loaded_model = FrankenPotential.load(model_save_path, map_location=device)

            # Step 4: Compare rf.state_dict between the original and loaded models
            assert are_dicts_close(
                model.rf.state_dict(), loaded_model.rf.state_dict(), verbose=True
            ), "The rf.state_dict() of the loaded model does not match the original model."
            assert are_dicts_close(
                model.input_scaler.state_dict(),
                loaded_model.input_scaler.state_dict(),
                verbose=True,
            ), "The input_scaler.state_dict() of the loaded model does not match the original model."

            # Step 5: Compare the hyperparameters between the original and loaded models
            assert (
                model.hyperparameters == loaded_model.hyperparameters
            ), "The hyperparameters of the loaded model do not match the original model."
        finally:
            if temp_dir is not None:
                cleanup_dir(temp_dir)


@pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
@pytest.mark.parametrize("device", DEVICES)
@pytest.mark.parametrize("scale_by_Z", [True, False])
def test_multiweight_save_load_functionality(rf_type, device, scale_by_Z):
    """Test for checking save and load methods of FrankenPotential"""
    for gnn_id in DEFAULT_GNN_IDS:
        temp_dir = None
        try:
            # Step 1: Create a temporary directory for saving the model
            temp_dir = create_temp_dir()

            data_path = DATASETS_PATH / DATASET_REGISTRY["test"]["test"]
            dataset = BaseAtomsDataset.from_path(
                data_path=data_path,
                split="train",
                gnn_backbone_id=gnn_id,
            )

            model = FrankenPotential(
                gnn_id,
                rf_type,
                random_features_params={},
                scale_by_Z=scale_by_Z,
                num_species=dataset.num_species,
            ).to(device)

            with torch.no_grad():
                gnn_features_stats = Statistics()
                for data, _ in dataset:  # type: ignore
                    data = data.to(device=device)
                    gnn_features = model.gnn.descriptors(data)
                    gnn_features_stats.update(
                        gnn_features, atomic_numbers=data.atomic_numbers
                    )

                model.input_scaler.set_from_statistics(gnn_features_stats)
                garbage_collection_cuda()

            multi_weights = torch.randn(
                17, model.rf.total_random_features, device=device, dtype=torch.float64
            )

            # Step 2: Save the model to the temporary directory
            model_save_path = os.path.join(temp_dir, "model_checkpoint.pth")
            model.save(model_save_path, multi_weights)

            # Step 3: Load the model from the saved checkpoint
            loaded_model = FrankenPotential.load(
                model_save_path,
                map_location=device,
                rf_weight_id=10,
            )

            model.rf.weights.copy_(multi_weights[10].reshape_as(model.rf.weights))

            # Step 4: Compare rf.state_dict between the original and loaded models
            assert are_dicts_close(
                model.rf.state_dict(), loaded_model.rf.state_dict(), verbose=True
            ), "The rf.state_dict() of the loaded model does not match the original model."

            # Step 5: Compare the hyperparameters between the original and loaded models
            assert (
                model.hyperparameters == loaded_model.hyperparameters
            ), "The hyperparameters of the loaded model do not match the original model."
        finally:
            if temp_dir is not None:
                cleanup_dir(temp_dir)


@pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
@pytest.mark.parametrize("device", DEVICES)
@pytest.mark.parametrize("multiweights", [True, False])
def test_inference_force_mode(rf_type, device, multiweights: bool):
    for gnn_id in DEFAULT_GNN_IDS:
        data_path = DATASETS_PATH / DATASET_REGISTRY["test"]["test"]
        dataset = BaseAtomsDataset.from_path(
            data_path=data_path,
            split="train",
            gnn_backbone_id=gnn_id,
            num_random_subsamples=1,
        )

        model = FrankenPotential(
            gnn_id,
            rf_type,
            random_features_params={},
            scale_by_Z=True,
            num_species=dataset.num_species,
        ).to(device)

        if multiweights:
            dummy_multiweights = torch.randn(
                (10,) + model.rf.weights.shape,
                dtype=model.rf.weights.dtype,
                device=model.rf.weights.device,
            )
            model.rf.weights = dummy_multiweights
        else:
            model.rf.weights.copy_(torch.randn(model.rf.weights.shape))

        calc_func = FrankenCalculator(model, device=device, forces_mode="torch.func")
        calc_autograd = FrankenCalculator(
            model,
            device=device,
            forces_mode="torch.autograd",
        )
        for atoms in dataset.ase_atoms[:1]:
            calc_func.calculate(atoms)
            calc_autograd.calculate(atoms)
            print(
                np.max(
                    np.abs(
                        calc_func.results["forces"] - calc_autograd.results["forces"]
                    )
                )
            )
            assert np.allclose(
                calc_func.results["forces"],
                calc_autograd.results["forces"],
                rtol=1e-3,
                atol=1e-3,
            )
            assert np.allclose(
                calc_func.results["energy"],
                calc_autograd.results["energy"],
                rtol=1e-3,
                atol=1e-3,
            )


@pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
@pytest.mark.parametrize("device", DEVICES)
@pytest.mark.parametrize("multiweights", [True, False])
class TestModelGradients:
    def test_gradients_mocked(self, rf_type, device, multiweights):
        num_atoms = 10
        dtype = torch.float32
        with mocked_gnn(device=device, dtype=dtype, feature_dim=32):
            model = FrankenPotential(
                "test",
                rf_type,
                random_features_params={},
            ).to(device)

        num_lin_models = 10 if multiweights else 1
        weights = torch.randn(
            (num_lin_models, model.rf.total_random_features), device=device
        )
        data = Configuration(
            torch.randn(num_atoms, 3, dtype=dtype),
            torch.randint(1, 100, (num_atoms,)),
            torch.tensor(num_atoms),
        ).to(device)
        energy_autograd, forces_autograd = model.energy_and_forces(
            data, weights=weights, forces_mode="torch.autograd"
        )
        energy_func, forces_func = model.energy_and_forces(
            data, weights=weights, forces_mode="torch.func"
        )
        torch.testing.assert_close(energy_func, energy_autograd, rtol=1e-3, atol=1e-3)
        torch.testing.assert_close(forces_func, forces_autograd, rtol=1e-3, atol=1e-3)

    @pytest.mark.parametrize("gnn_id", DEFAULT_GNN_IDS)
    def test_gradients_real(self, rf_type, device, multiweights: bool, gnn_id):
        # Define the rng_seed and initialize the model
        model = FrankenPotential(
            gnn_id,
            rf_type,
            random_features_params={},
        ).to(device)

        num_lin_models = 10 if multiweights else 1
        weights = torch.randn(
            (num_lin_models, model.rf.total_random_features), device=device
        )
        data_path = DATASETS_PATH / DATASET_REGISTRY["test"]["test"]
        dataset = BaseAtomsDataset.from_path(
            data_path=data_path,
            split="train",
            gnn_backbone_id=gnn_id,
        )
        data, _ = dataset[0]  # type: ignore
        data = data.to(device)
        energy_autograd, forces_autograd = model.energy_and_forces(
            data, weights=weights, forces_mode="torch.autograd"
        )
        energy_func, forces_func = model.energy_and_forces(
            data, weights=weights, forces_mode="torch.func"
        )
        torch.testing.assert_close(energy_func, energy_autograd, rtol=1e-3, atol=1e-3)
        torch.testing.assert_close(forces_func, forces_autograd, rtol=1e-3, atol=1e-3)


@pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
@pytest.mark.parametrize("device", DEVICES)
@pytest.mark.parametrize("gnn_id", DEFAULT_GNN_IDS)
def test_calculator_in_md(rf_type, device, gnn_id):
    # Define the rng_seed and initialize the model
    model = FrankenPotential(
        gnn_id,
        rf_type,
        random_features_params={},
    ).to(device)
    num_lin_models = 1  # only a single weight for MD
    rf_weights = torch.randn(
        (num_lin_models, model.rf.total_random_features), device=device
    )
    model.rf.weights = rf_weights
    calculator = FrankenCalculator(model, device=device, forces_mode="torch.autograd")

    # Molecular dynamics
    # 1. Get the initial configuration
    # 2. Set some attribute on the configuration with MaxwellBoltzmannDistribution
    # 3. Create and run the MD
    data_path = DATASETS_PATH / DATASET_REGISTRY["test"]["md"]

    init_traj_atoms = read(data_path, index=0)
    assert isinstance(init_traj_atoms, ase.Atoms)
    init_traj_atoms.calc = calculator
    MaxwellBoltzmannDistribution(init_traj_atoms, temperature_K=500)
    md = NPT(
        init_traj_atoms,
        timestep=1 * units.fs,
        temperature_K=500,
        ttime=25 * units.fs,
        logfile="-",
        trajectory=None,
        loginterval=1,
    )
    md.run(2)


@pytest.mark.parametrize("rf_type", RF_PARAMETRIZE)
@pytest.mark.parametrize("gnn_id", DEFAULT_GNN_IDS)
@pytest.mark.parametrize("device", DEVICES)
@pytest.mark.parametrize("atomic_energies", [None, {7: 1.0, 26: 10.0}])
def test_autotune(rf_type, gnn_id, device, atomic_energies):
    loaders = init_loaders(
        gnn_id,
        DATASETS_PATH / DATASET_REGISTRY["test"]["train"],
        test_path=DATASETS_PATH / DATASET_REGISTRY["test"]["test"],
        val_path=DATASETS_PATH / DATASET_REGISTRY["test"]["val"],
    )
    trainer = RandomFeaturesTrainer(
        train_dataloader=loaders["train"],
        random_features_normalization=None,
        log_dir=None,  # TODO: Test with log_dir not None (but mocked) since there was a regression here
        save_every_model=False,
        device=device,
    )
    rf_hp_grid = [
        (
            0,
            {
                "num_random_features": 256,
                "length_scale": 1.0,
            },
        ),
        (
            1,
            {
                "num_random_features": 512,
                "length_scale": 1.0,
            },
        ),
    ]
    solver_hp_grid = {
        "L2_penalty": [1e-8],
        "loss_lerp_weight": [0.1, 0.9],
    }
    best_model = run_autotune(
        rnd_seed=42,
        gnn_backbone_id=gnn_id,
        interaction_block=2,
        rf_type=rf_type,
        loaders=loaders,
        rf_param_grid=rf_hp_grid,
        solver_param_grid=solver_hp_grid,
        trainer=trainer,
        scale_by_Z=False,
        atomic_energies=atomic_energies,
        jac_chunk_size="auto",
    )
    print(best_model)
    # TODO: Assert something to prevent regressions


class TestStatistics:
    def test_online_algo(self):
        dim = 128
        atomic_numbers = torch.tensor([1, 5, 1])
        data = torch.randn(100, len(atomic_numbers), dim)
        st = Statistics(input_dim=dim)

        for data_item in data:
            st.update(data_item, atomic_numbers)

        # Compute the expected values
        global_mean = data.view(-1, dim).mean(0)
        global_std = data.view(-1, dim).std(0)

        torch.testing.assert_close(st.statistics[0]["mean"], global_mean.double())
        torch.testing.assert_close(
            st.statistics[0]["std"], global_std.double(), rtol=1e-2, atol=1e-2
        )

    def test_per_atom(self):
        dim = 128
        atomic_numbers = torch.tensor([1, 5, 1])
        data = torch.randn(100, len(atomic_numbers), dim)
        st = Statistics(input_dim=dim)

        for data_item in data:
            st.update(data_item, atomic_numbers)

        # Compute the expected values
        atom1_mean = data[:, [0, 2]].view(-1, dim).mean(0)
        atom5_mean = data[:, [1]].view(-1, dim).mean(0)
        atom1_std = data[:, [0, 2]].view(-1, dim).std(0)
        atom5_std = data[:, [1]].view(-1, dim).std(0)

        torch.testing.assert_close(st.statistics[1]["mean"], atom1_mean.double())
        torch.testing.assert_close(
            st.statistics[1]["std"], atom1_std.double(), rtol=1e-2, atol=1e-2
        )
        torch.testing.assert_close(st.statistics[5]["mean"], atom5_mean.double())
        torch.testing.assert_close(
            st.statistics[5]["std"], atom5_std.double(), rtol=1e-2, atol=1e-2
        )


class TestFit:
    pass


class TestSerialize:
    pass



import hashlib
import logging
import math
from pathlib import Path
from time import perf_counter
from typing import List, Literal, Optional, Sequence, Union

import numpy as np
import torch
import torch.utils.data
from torch import Tensor

import franken.metrics
from franken.metrics.base import BaseMetric
import franken.utils.distributed as dist_utils
from franken.data.base import Target
from franken.rf.model import FrankenPotential
from franken.trainers import BaseTrainer
from franken.trainers.log_utils import (
    DataSplit,
    HyperParameterGroup,
    LogCollection,
    LogEntry,
)
from franken.utils.linalg.cov import (
    lowmem_normalize_leading_eig,
    rank1_update,
    rankk_update,
)
from franken.utils.linalg.psdsolve import psd_ridge
from franken.utils.linalg.tri import triangular_lerp
from franken.utils.misc import params_grid, throughput


logger = logging.getLogger("franken")


class RandomFeaturesTrainer(BaseTrainer):
    def __init__(
        self,
        train_dataloader: torch.utils.data.DataLoader,
        random_features_normalization: Optional[Literal["leading_eig"]] = "leading_eig",
        log_dir: Path | None = None,  # If None, logging is disabled
        save_every_model: bool = True,
        device: Union[torch.device, str, int] = "cuda:0",
        dtype: Union[str, torch.dtype] = torch.float32,
        save_fmaps: bool = True,
    ):
        super().__init__(
            train_dataloader,
            log_dir=log_dir,
            save_every_model=save_every_model,
            device=device,
            dtype=dtype,
        )
        self.random_features_normalization = random_features_normalization
        self.save_fmaps = save_fmaps

    def on_fit_start(self, model: FrankenPotential):
        # initialize input scaler based on statistics property
        model.input_scaler.set_from_statistics(self.get_statistics(model)[0])
        # initialize energy shift based on atomic energies
        if not model.energy_shift.is_initialized:
            model.energy_shift.set_from_atomic_energies(
                self.train_dataloader.dataset.atomic_energies
            )

    def fit(
        self,
        model: FrankenPotential,
        solver_params: dict,
    ) -> tuple[LogCollection, torch.Tensor]:
        model = model.to(self.device)
        self.on_fit_start(model)
        model_hash = hashlib.md5(str(model.hyperparameters).encode())
        model_hash = model_hash.hexdigest()

        t_cov_coeffs_start = perf_counter()
        self._compute_covs_and_coeffs(model, self.train_dataloader)
        t_cov_coeffs = perf_counter() - t_cov_coeffs_start

        solver_grid_size = math.prod([len(v) for v in solver_params.values()])
        all_weights = torch.zeros(
            solver_grid_size,
            model.rf.total_random_features,
            dtype=self.buffer_dt,
            device=self.device,
        )

        solver_iter = throughput(
            params_grid(solver_params, split_distributed=True),
            desc="least-squares",
            units="models",
            device=self.device,
        )

        local_logs = dict()
        num_failed = torch.zeros((1,), device=self.device, dtype=torch.int)
        for hp_idx, hp_val in solver_iter:
            t_solve_start = perf_counter()
            try:
                weights = self._solve(**hp_val)
            except torch.linalg.LinAlgError as e:
                weights = torch.full_like(all_weights[hp_idx], torch.inf)
                num_failed += 1
                logger.debug(f"Hyperparameter {hp_val} failed. Error: {e}")
            finally:
                t_solve = perf_counter() - t_solve_start
            # Update all weights
            all_weights[hp_idx].copy_(weights.view(-1))

            # Logging
            solver_hps = hp_val | {"dtype": self.buffer_dt}
            hp_groups = model.hyperparameters | {"solver": solver_hps}
            hyperparameters = []
            for group_name, hps in hp_groups.items():
                hyperparameters.append(HyperParameterGroup.from_dict(group_name, hps))

            local_logs[hp_idx] = LogEntry(
                checkpoint_hash=model_hash,
                checkpoint_rf_weight_id=hp_idx,
                timings_cov_coeffs=t_cov_coeffs,
                timings_solve=t_solve,
                hyperparameters=hyperparameters,
            )

        # Broadcast weights and logs
        dist_utils.all_sum(all_weights)
        log_collection = LogCollection.gather_from_ranks(local_logs)
        assert len(log_collection) == solver_grid_size

        # Reporting failed runs
        dist_utils.all_sum(num_failed)

        if num_failed.item() > 0:
            logger.warning(
                f"Solver failed in {num_failed.item()}/{solver_grid_size} cases."
            )

        return log_collection, all_weights

    def evaluate(
        self,
        model: FrankenPotential,
        dataloader: torch.utils.data.DataLoader,
        log_collection: LogCollection,
        all_weights: torch.Tensor | None,
        metrics: Sequence[str] = ("energy_MAE", "forces_MAE", "forces_cosim"),
    ) -> LogCollection:
        tot_dset_size = len(dataloader.dataset)  # type: ignore

        metric_objects: List[BaseMetric] = []
        for name in metrics:
            try:
                metric_obj = franken.metrics.init_metric(
                    name, device=self.device, dtype=self.buffer_dt
                )
                metric_objects.append(metric_obj)
            except KeyError:
                logger.warning(
                    f"Unknown metric {name}. Skipping. Available metrics: {franken.metrics.available_metrics()}"
                )

        split_name = dataloader.dataset.split
        try:
            split = DataSplit[split_name.upper()]
        except KeyError:
            logger.warning(f"Unrecognized split '{split_name}' in dataloader")
            split = DataSplit.UNDEFINED

        progress_bar = throughput(
            dataloader,
            desc=f"{split.name.lower()} evaluation",
            total=tot_dset_size,
            device=self.device,
        )
        with torch.no_grad():
            for i, (data, targets) in enumerate(progress_bar):
                data = data.to(device=self.device)
                targets = targets.to(device=self.device)
                if split == DataSplit.TRAIN and self.save_fmaps:
                    # Shortcut to compute predictions for the training-set, for which
                    # we already have computed energy and force feature maps
                    assert len(self.forces_fmap) == len(self.energy_fmap)
                    assert len(self.forces_fmap) == len(dataloader)
                    predictions = Target(
                        *model.energy_and_forces_from_fmaps(
                            data,
                            energy_fmap=self.energy_fmap[i],
                            forces_fmap=self.forces_fmap[i],
                            weights=all_weights,
                            add_energy_shift=False,  # since it's always train here
                        )
                    )
                else:
                    predictions = Target(
                        *model.energy_and_forces(
                            data,
                            weights=all_weights,
                            forces_mode="torch.func",
                            add_energy_shift=(
                                False if split == DataSplit.TRAIN else True
                            ),
                        )
                    )
                for metric in metric_objects:
                    metric.update(predictions, targets)

        num_models = (
            all_weights.shape[0]
            if all_weights is not None
            else model.rf.weights.shape[0]
        )
        # Sync metrics across GPUs
        metric_values = {}
        for metric in metric_objects:
            value = metric.compute(reset=False)
            if metric.samples_counter.item() != tot_dset_size:
                logger.warning(
                    f"Metric {metric.name} has {metric.samples_counter.item()} samples "
                    f"while the dataset has {tot_dset_size} configurations."
                )
            assert value.ndim == 1
            assert value.shape[0] == num_models
            metric_values[metric.name] = value

        # Explode metric values into list of MetricLog
        raw_logs = []
        for idx in range(num_models):
            results = []
            for name, value in metric_values.items():
                float_value = value[idx].item()
                results.append(dict(name=name, value=float_value))
            raw_logs.append(results)

        assert len(raw_logs) == len(log_collection)

        for log_entry, results in zip(log_collection, raw_logs):
            for metric in results:
                log_entry.add_metric(metric["name"], metric["value"], split)
        return log_collection

    def warn_save_fmaps(
        self, energy_fmap: Tensor, forces_fmap: Tensor, num_maps: int
    ) -> None:
        fmap_size = np.prod(energy_fmap.shape) + np.prod(forces_fmap.shape)
        tot_bytes = fmap_size * forces_fmap.element_size() * num_maps
        if self.device.type == "cuda":
            avail_bytes = torch.cuda.mem_get_info(self.device)[0]
            if tot_bytes > 0.8 * avail_bytes:
                logger.warning(
                    f"Saved feature maps require {tot_bytes / 2**30:.2f}GB of device memory. "
                    f"Device has {avail_bytes / 2**30:.2f}GB of available memory, a crash may occur. "
                    f"Use 'trainer.save_fmaps=False' in your config to stop saving feature maps."
                )
            else:
                logger.info(
                    f"Saved feature maps require {tot_bytes / 2**30:.2f}GB of device memory."
                )

    @torch.no_grad()
    def _compute_covs_and_coeffs(
        self, model: FrankenPotential, dataloader: torch.utils.data.DataLoader
    ) -> None:
        tot_dset_size = len(dataloader.dataset)  # type: ignore
        n_rf = model.rf.total_random_features
        # Initialize Buffers: `covariance` will have
        # - covariance of forces on LOWER triangle
        # - covariance of energies on UPPER triangle
        # with diagonals stored separately, and linsys coefficients.
        self.covariance = torch.zeros(
            (n_rf, n_rf), device=self.device, dtype=self.buffer_dt
        )
        self.diag_energy = torch.zeros(
            (n_rf,), device=self.device, dtype=self.buffer_dt
        )
        self.diag_forces = torch.zeros(
            (n_rf,), device=self.device, dtype=self.buffer_dt
        )
        self.coeffs_energy = torch.zeros(
            (n_rf,), device=self.device, dtype=self.buffer_dt
        )
        self.coeffs_forces = torch.zeros(
            (n_rf,), device=self.device, dtype=self.buffer_dt
        )
        # NOTE: the feature maps are never synced between devices.
        #       They can only used correctly by iterating through the
        #       same dataloader as here, using the same method. Otherwise
        #       they may not be in the correct order
        self.energy_fmap = []
        self.forces_fmap = []

        progress_bar = throughput(
            dataloader,
            desc="covs+coeffs",
            total=tot_dset_size,
            device=self.device,
        )

        for i, (data, targets) in enumerate(progress_bar):
            data = data.to(device=self.device)
            targets = targets.to(device=self.device)

            energy_per_atom = targets.energy / data.natoms
            forces_per_atom = targets.forces / data.natoms

            forces_fmap, energy_fmap = model.grad_feature_map(data)
            energy_fmap = energy_fmap.to(dtype=self.buffer_dt)

            rank1_update(self.covariance, self.diag_energy, energy_fmap, upper=True)
            self.coeffs_energy.add_(energy_fmap, alpha=energy_per_atom.item())

            forces_fmap = (-forces_fmap).to(dtype=self.buffer_dt)
            forces_fmap = forces_fmap.view(forces_fmap.shape[0], -1)
            rankk_update(self.covariance, self.diag_forces, forces_fmap, upper=False)
            self.coeffs_forces.addmv_(
                forces_fmap, forces_per_atom.view(-1).to(dtype=self.buffer_dt)
            )
            if self.save_fmaps:
                if i == 0:
                    self.warn_save_fmaps(energy_fmap, forces_fmap, len(dataloader))
                self.energy_fmap.append(energy_fmap)
                self.forces_fmap.append(forces_fmap)

        # Sync covariance matrices & coefficients
        dist_utils.all_sum(self.covariance)
        dist_utils.all_sum(self.diag_forces)
        dist_utils.all_sum(self.diag_energy)
        dist_utils.all_sum(self.coeffs_energy)
        dist_utils.all_sum(self.coeffs_forces)

        if self.random_features_normalization == "leading_eig":
            logger.warning(
                "`leading_eig` normalization has high memory usage. If you encounter OOM errors try to disable it."
            )
            lowmem_normalize_leading_eig(
                self.covariance, self.diag_energy, self.coeffs_energy, upper=True
            )
            lowmem_normalize_leading_eig(
                self.covariance, self.diag_forces, self.coeffs_forces, upper=False
            )
        elif self.random_features_normalization is not None:
            raise ValueError(
                f"Covariance normalization {self.random_features_normalization} is not implemented."
            )

    @torch.no_grad()
    def _solve(self, loss_lerp_weight: float, L2_penalty: float = 1e-6) -> Tensor:
        # This is the 2nd copy of the covariance matrix that we need to store.
        lerped_cov, lerped_diag = triangular_lerp(
            self.covariance,
            diag_upper=self.diag_energy,
            diag_lower=self.diag_forces,
            weight=loss_lerp_weight,
            inplace=False,
        )
        lerped_cov.diagonal().copy_(lerped_diag)
        rhs = torch.lerp(self.coeffs_energy, self.coeffs_forces, loss_lerp_weight)
        solution = psd_ridge(lerped_cov, rhs, L2_penalty)
        return solution



from pathlib import Path
from types import MethodType
from typing import Union
from functools import partial

import sevenn._keys as KEY
import torch
import torch.nn as nn
from sevenn.util import model_from_checkpoint

from franken.data.base import Configuration


def extract_scalar_irrep(data, irreps):
    node_features = data[KEY.NODE_FEATURE]
    scalar_slice = irreps.slices()[0]
    scalar_features = node_features[..., scalar_slice]
    return scalar_features


def franken_sevenn_descriptors(
    self,
    data: Configuration,
    interaction_layer: int,
    extract_after_act: bool = True,
    append_layers: bool = True,
):
    # Convert data to sevenn
    assert data.cell is not None
    sevenn_data = {
        KEY.NODE_FEATURE: data.atomic_numbers,
        KEY.ATOMIC_NUMBERS: data.atomic_numbers,
        KEY.POS: data.atom_pos,
        KEY.EDGE_IDX: data.edge_index,
        KEY.CELL: data.cell,
        KEY.CELL_SHIFT: data.shifts,  # TODO: Check this correct?
        KEY.CELL_VOLUME: torch.einsum(
            "i,i", data.cell[0, :], torch.linalg.cross(data.cell[1, :], data.cell[2, :])
        ),
        KEY.NUM_ATOMS: len(data.atomic_numbers),
        KEY.BATCH: data.batch_ids,
    }

    # From v0.9.3 to v10 sevenn introduced some changes in how models are build
    # (`build_E3_equivariant_model`), removing the EdgePreprocess class before the
    # network itself. The main purpose of EdgePreprocess was to initialize the
    # KEY.EDGE_VEC (r_ij: the vector between atom positions) and KEY.EDGE_LENGTH.
    # We replace that functionality here.
    # NOTE: the original preprocess had some special handling of the PBC cell
    #       when self.is_stress was set to True. We're ignoring all that.
    # NOTE: as comparison to the original EdgePreprocess we assume `is_batch_data`
    #       to be False.
    idx_src = sevenn_data[KEY.EDGE_IDX][0]
    idx_dst = sevenn_data[KEY.EDGE_IDX][1]
    pos = sevenn_data[KEY.POS]
    edge_vec = pos[idx_dst] - pos[idx_src]
    edge_vec = edge_vec + torch.einsum(
        "ni,ij->nj", sevenn_data[KEY.CELL_SHIFT], sevenn_data[KEY.CELL].view(3, 3)
    )
    sevenn_data[KEY.EDGE_VEC] = edge_vec
    sevenn_data[KEY.EDGE_LENGTH] = torch.linalg.norm(edge_vec, dim=-1)

    # Iterate through the model's layers
    # the sanest way to figure out which layer we're at is through the
    # `_modules` attribute of `nn.Sequential` (which the Sevenn network
    # inherits from), which exposes key-value pairs.
    layer_idx = 0
    scalar_features_list = []
    for i, (name, module) in enumerate(self._modules.items()):
        if "self_connection_intro" in name:
            layer_idx += 1

        new_sevenn_data = module(sevenn_data)
        if "equivariant_gate" in name:
            if extract_after_act:
                scalar_features = extract_scalar_irrep(
                    new_sevenn_data, module.gate.irreps_out
                )
            else:
                scalar_features = extract_scalar_irrep(
                    sevenn_data, module.gate.irreps_in
                )
            if append_layers:
                scalar_features_list.append(scalar_features)
            else:
                scalar_features_list[0] = scalar_features
            if layer_idx == interaction_layer:
                break
        sevenn_data = new_sevenn_data

    return torch.cat(scalar_features_list, dim=-1)


def franken_sevenn_num_params(self) -> int:
    return sum(p.numel() for p in self.parameters())


def franken_sevenn_feature_dim(
    self,
    interaction_layer: int,
    extract_after_act: bool = True,
    append_layers: bool = True,
):
    layer_idx = 0
    tot_feat_dim = 0
    for i, (name, module) in enumerate(self._modules.items()):
        if "self_connection_intro" in name:
            layer_idx += 1
        if "equivariant_gate" in name:
            if extract_after_act:
                new_feat_dim = module.gate.irreps_out.count("0e")
            else:
                new_feat_dim = module.gate.irreps_in.count("0e")
            if append_layers:
                tot_feat_dim += new_feat_dim
            else:
                tot_feat_dim = new_feat_dim
            if layer_idx == interaction_layer:
                break
    return tot_feat_dim


class FrankenSevenn:
    @staticmethod
    def load_from_checkpoint(
        trainer_ckpt: Union[str, Path],
        gnn_backbone_id: str,
        interaction_block: int,
        extract_after_act: bool = True,
        append_layers: bool = True,
    ):
        sevenn, config = model_from_checkpoint(str(trainer_ckpt))
        assert isinstance(sevenn, nn.Module)
        sevenn.descriptors = MethodType(  # type: ignore
            partial(
                franken_sevenn_descriptors,
                interaction_layer=interaction_block,
                extract_after_act=extract_after_act,
                append_layers=append_layers,
            ),
            sevenn,
        )
        sevenn.num_params = MethodType(franken_sevenn_num_params, sevenn)  # type: ignore
        sevenn.feature_dim = MethodType(  # type: ignore
            partial(
                franken_sevenn_feature_dim,
                interaction_layer=interaction_block,
                extract_after_act=extract_after_act,
                append_layers=append_layers,
            ),
            sevenn,
        )

        def init_args(self):
            return {
                "gnn_backbone_id": gnn_backbone_id,
                "interaction_block": interaction_block,
                "extract_after_act": extract_after_act,
                "append_layers": append_layers,
            }

        sevenn.init_args = MethodType(init_args, sevenn)  # type: ignore
        return sevenn
